#!/usr/bin/python3
"""This script will block until the specified file appears or the timeout
   expires. The first argument is the timeout in seconds. The second argument
   is the filepath."""
import os
import time
import sys


def logging(*args):
    """add a prefix to print() output"""
    print("wait_for_warmup:", ' '.join(args), flush=True)


def wait_for_warmup(warmup_file, max_wait_sec):
    """wait for warmup_file to appear until max_wait_sec timeout"""
    counter = 0
    msg = warmup_file
    while True:
        if os.access(warmup_file, os.F_OK):
            msg += " is detected!"
            break
        if counter >= max_wait_sec:
            msg += " didn't show up!"
            break
        counter += 1
        time.sleep(1)
    msg += " Time elapsed = %d" % counter
    logging(msg)


def main():
    """Program starts here"""
    warmup_file = '/tmp/pipeline.warmup.done'
    max_wait_sec = 60

    for i in range(len(sys.argv)):
        arg = sys.argv[i]
        if i == 1:
            max_wait_sec = int(arg)
        elif i == 2:
            warmup_file = arg
    if max_wait_sec > 0:
        wait_for_warmup(warmup_file, max_wait_sec)


if __name__ == '__main__':
    main()



#!/usr/bin/python

import subprocess

from Tkinter import *

class Application(Frame):
    def say_hi(self):
        print "hi there, everyone!"

    def publishSnapshotMessage(self):
        cmd = 'mosquitto_pub -h localhost -t berrynet/event/camera -m snapshot_picam'
        subprocess.call(cmd, shell=True)

    def createWidgets(self):
        self.hi_there = Button(self, text='Snapshot', width=16, height=5)
        self.hi_there["command"] = self.publishSnapshotMessage
        self.hi_there.pack({"side": "left"})

    def __init__(self, master=None):
        Frame.__init__(self, master)
        self.pack()
        self.createWidgets()

root = Tk()
app = Application(master=root)
app.master.title('Cheese')
app.master.lift()
app.master.attributes('-topmost', True)
app.mainloop()
root.destroy()



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Simple image classification server with Inception.

The server monitors image_dir and run inferences on new images added to the
directory. Every image file should come with another empty file with '.done'
suffix to signal readiness. Inference result of a image can be read from the
'.txt' file of that image after '.txt.done' is spotted.

This is an example the server expects clients to do. Note the order.

# cp cat.jpg /run/image_dir
# touch /run/image_dir/cat.jpg.done

Clients should wait for appearance of 'cat.jpg.txt.done' before getting
result from 'cat.jpg.txt'.
"""


from __future__ import print_function

import argparse
import os
import sys
import time

from datetime import datetime

import cv2
import movidius as mv
import numpy as np

mvng = None


def logging(*args):
    print("[%08.3f]" % time.time(), ' '.join(args))


def run(mvng):
    """Infinite loop serving inference requests"""

    logging('detection service is running')

    capture = cv2.VideoCapture(0)
    while True:
        t_start = datetime.now()
        status, image_data = capture.read()
        t_end = datetime.now()
        t_capture = t_end - t_start
        logging('capture time: {} ms'.format(t_capture.total_seconds() * 1000))

        t_start = datetime.now()
        processed_data = mv.process_yolo_input(image_data)
        t_end = datetime.now()
        t_preprocess = t_end - t_start
        logging('preprocess time: {} ms'.format(t_preprocess.total_seconds() * 1000))

        t_start = datetime.now()
        output = mvng.inference(processed_data)
        t_end = datetime.now()
        t_inference = t_end - t_start
        logging('inference time: {} ms'.format(t_inference.total_seconds() * 1000))

        t_start = datetime.now()
        interpreted_output = mv.interpret_yolo_output(output,
                                                      image_data.shape[1],
                                                      image_data.shape[0])
        yolo_outputs = mv.process_yolo_output(image_data,
                                              interpreted_output,
                                              '/tmp/yolo_resutl.jpg')
        t_end = datetime.now()
        t_postprocess = t_end - t_start
        logging('postprocess time: {} ms'.format(t_postprocess.total_seconds() * 1000))

        #output_name = input_name + '.txt'
        #mv.save_yolo_output_text(output_name, yolo_outputs)
        #logging(input_name, " detected!")


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--model', required=True, help='Model file path')
    ap.add_argument('--label', required=True, help='Label file path')
    ap.add_argument('--image_dir', required=True, help='Path to image file')
    #ap.add_argument('--num_top_predictions', default=5,
    #                help='Display this many predictions')
    return vars(ap.parse_args())


if __name__ == '__main__':
    args = parse_args()

    mvng = mv.MovidiusNeuralGraph(args['model'], args['label'])

    logging("model filepath: ", args['model'])
    logging("label filepath: ", args['label'])
    logging("image_dir: ", args['image_dir'])

    # workaround the issue that SIGINT cannot be received (fork a child to
    # avoid blocking the main process in Thread.join()
    child_pid = os.fork()
    if child_pid == 0:  # child
        run(mvng)
    else:  # parent
        try:
            os.wait()
        except KeyboardInterrupt:
            os.kill(child_pid, signal.SIGKILL)
            os.unlink(pidfile)



#!/usr/bin/python

import subprocess

from Tkinter import *

class Application(Frame):
    def publishSnapshotMessage(self):
        cmd = 'mosquitto_pub -h localhost -t berrynet/event/camera -m snapshot_picam'
        subprocess.call(cmd, shell=True)

    def startGateway(self):
        cmd = 'bash ~/codes/BerryNet/berrynet-manager start'
        subprocess.call(cmd, shell=True)

    def stopGateway(self):
        cmd = 'bash ~/codes/BerryNet/berrynet-manager stop'
        subprocess.call(cmd, shell=True)

    def startChromium(self):
        cmd = 'bash -c "sensible-browser http://localhost:8080/index.html#source=dashboard.json" &'
        subprocess.call(cmd, shell=True)

    def stopChromium(self):
        cmd = 'bash -c "kill -9 $(pgrep chromium)"'
        subprocess.call(cmd, shell=True)

    def startDemo(self):
        self.startGateway()
        self.startChromium()

    def stopDemo(self):
        self.stopChromium()
        self.stopGateway()

    def cleanSnapshots(self):
        cmd = 'bash -c "sudo rm /usr/local/berrynet/inference/image/snapshot*"'
        subprocess.call(cmd, shell=True)

    def createWidgets(self):
        self.snapshotButton = Button(self, text='Snapshot', width=16, height=5)
        self.snapshotButton["command"] = self.publishSnapshotMessage
        self.snapshotButton.pack(fill=X)

        self.startDemoButton = Button(self, text='Start Demo', width=16, height=5)
        self.startDemoButton["command"] = self.startDemo
        self.startDemoButton.pack(fill=X)

        self.stopDemoButton = Button(self, text='Stop Demo', width=16, height=5)
        self.stopDemoButton["command"] = self.stopDemo
        self.stopDemoButton.pack(fill=X)

        self.cleanButton = Button(self, text='Clean Snapshots', width=16, height=5)
        self.cleanButton["command"] = self.cleanSnapshots
        self.cleanButton.pack(fill=X)

    def __init__(self, master=None):
        Frame.__init__(self, master)
        self.pack()
        self.createWidgets()

root = Tk()
app = Application(master=root)
app.master.title('Cheese')
app.master.lift()
app.master.attributes('-topmost', True)
app.mainloop()
root.destroy()



import os

from logzero import setup_logger


# Save log file at different place to prevent permission error.
if os.geteuid() == 0:  # root
    LOGGING_FLLEPATH='/tmp/berrynet.log'
else:
    LOGGING_FLLEPATH='{}/.cache/berrynet.log'.format(os.getenv('HOME'))

logger = setup_logger(name='berrynet-logger', logfile=LOGGING_FLLEPATH)



import logging
import time

from argparse import ArgumentParser
from os import path

import cv2
import numpy as np
import tensorflow as tf

from berrynet.engine import DLEngine
from berrynet import logger


class TFLiteDetectorEngine(DLEngine):
    def __init__(self, model, labels, threshold=0.5, num_threads=1):
        """
        Builds Tensorflow graph, load model and labels
        """
        # Load labels
        self.labels = self._load_label(labels)
        self.classes = len(self.labels)

        # Define lite graph and Load Tensorflow Lite model into memory
        self.interpreter = tf.lite.Interpreter(
            model_path=model,
            num_threads=num_threads)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        self.input_dtype = self.input_details[0]['dtype']
        self.threshold = threshold

    def __delete__(self, instance):
        #tf.reset_default_graph()
        #self.sess = tf.InteractiveSession()
        del self.interpreter

    def process_input(self, tensor):
        """Resize and normalize image for network input"""

        self.img_w = tensor.shape[1]
        self.img_h = tensor.shape[0]

        frame = cv2.cvtColor(tensor, cv2.COLOR_BGR2RGB)
        frame = cv2.resize(frame, (300, 300))
        frame = np.expand_dims(frame, axis=0)
        if self.input_dtype == np.float32:
            frame = (2.0 / 255.0) * frame - 1.0
            frame = frame.astype('float32')
        else:
            # default data type returned by cv2.imread is np.unit8
            pass
        return frame

    def inference(self, tensor):
        self.interpreter.set_tensor(self.input_details[0]['index'], tensor)
        self.interpreter.invoke()

        # get results
        boxes = self.interpreter.get_tensor(
            self.output_details[0]['index'])
        classes = self.interpreter.get_tensor(
            self.output_details[1]['index'])
        scores = self.interpreter.get_tensor(
            self.output_details[2]['index'])
        num = self.interpreter.get_tensor(
            self.output_details[3]['index'])
        return {
            'boxes': boxes,
            'classes': classes,
            'scores': scores,
            'num': num
        }

    def process_output(self, output):
        # get results
        boxes = np.squeeze(output['boxes'][0])
        classes = np.squeeze(output['classes'][0] + 1).astype(np.int32)
        scores = np.squeeze(output['scores'][0])
        num = output['num'][0]

        annotations = []
        number_boxes = boxes.shape[0]
        for i in range(number_boxes):
            box = tuple(boxes[i].tolist())
            ymin, xmin, ymax, xmax = box

            if scores[i] < self.threshold:
                continue
            annotations.append({
                'label': self.labels[classes[i]],
                'confidence': float(scores[i]),
                'left': int(xmin * self.img_w),
                'top': int(ymin * self.img_h),
                'right': int(xmax * self.img_w),
                'bottom': int(ymax * self.img_h)
            })
        return {'annotations': annotations}

    def _load_label(self, path):
        with open(path, 'r') as f:
            labels = list(map(str.strip, f.readlines()))
        return labels


class TFLiteClassifierEngine(DLEngine):
    def __init__(self, model, labels, top_k=3, num_threads=1,
                 input_mean=127.5, input_std=127.5):
        """
        Builds Tensorflow graph, load model and labels
        """
        # Load labels
        self.labels = self._load_label(labels)
        self.classes = len(self.labels)

        # Define lite graph and Load Tensorflow Lite model into memory
        self.interpreter = tf.lite.Interpreter(
            model_path=model,
            num_threads=num_threads)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        self.floating_model = False
        if self.input_details[0]['dtype'] == np.float32:
            self.floating_model = True
        self.input_mean = input_mean
        self.input_std = input_std
        self.top_k = int(top_k)

    def __delete__(self, instance):
        #tf.reset_default_graph()
        #self.sess = tf.InteractiveSession()
        del self.interpreter

    def process_input(self, tensor):
        """Resize and normalize image for network input"""

        self.img_w = tensor.shape[1]
        self.img_h = tensor.shape[0]

        frame = cv2.cvtColor(tensor, cv2.COLOR_BGR2RGB)
        frame = cv2.resize(frame, (self.input_details[0]['shape'][2],
                                   self.input_details[0]['shape'][1]))
        frame = np.expand_dims(frame, axis=0)
        if self.floating_model:
            frame = (np.float32(frame) - self.input_mean) / self.input_std
        return frame

    def inference(self, tensor):
        self.interpreter.set_tensor(self.input_details[0]['index'], tensor)
        self.interpreter.invoke()
        output_data = self.interpreter.get_tensor(self.output_details[0]['index'])
        results = np.squeeze(output_data)
        return {
            'scores': results,
        }

    def process_output(self, output):
        # get results
        scores = output['scores']
        top_k_results = scores.argsort()[-self.top_k:][::-1]

        processed_output = {'annotations': []}

        for i in top_k_results:
            human_string = self.labels[i]
            if self.floating_model:
                score = float(scores[i])
            else:
                score = float(scores[i])/255.0
            anno = {
                'type': 'classification',
                'label': human_string,
                'confidence': score
            }
            processed_output['annotations'].append(anno)

        return processed_output

    def _load_label(self, path):
        with open(path, 'r') as f:
            labels = list(map(str.strip, f.readlines()))
        return labels


def parse_argsr():
    parser = ArgumentParser()
    parser.add_argument(
        "-e", "--engine",
        help=("Classifier or Detector engine. "
              "classifier, or detector is acceptable. "
              "(classifier by default)"),
        default="classifier",
        type=str)
    parser.add_argument(
        "-m", "--model",
        help="Path to an .xml file with a trained model.",
        required=True,
        type=str)
    parser.add_argument(
        "-l", "--labels",
        help="Labels mapping file",
        default=None,
        type=str)
    parser.add_argument(
        "--top_k",
        help="Number of top results",
        default=3,
        type=int)
    parser.add_argument(
        "--num_threads",
        help="Number of threads",
        default=1,
        type=int)
    parser.add_argument(
        "-i", "--input",
        help="Path to a folder with images or path to an image files",
        required=True,
        type=str)
    parser.add_argument(
        "--debug",
        help="Debug mode toggle",
        default=False,
        action="store_true")

    return parser.parse_args()

def main():
    # Example command
    #     $ python3 tflite_engine.py -e detector \
    #           -m detect.tflite --labels labels.txt -i dog.jpg --debug
    args = parse_argsr()

    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if args.engine == 'classifier':
        engine = TFLiteClassifierEngine(
                     model = args.model,
                     labels = args.labels,
                     top_k = args.top_k,
                     num_threads = args.num_threads)
    elif args.engine == 'detector':
        engine = TFLiteDetectorEngine(
                     model = args.model,
                     labels = args.labels,
                     num_threads = args.num_threads)
    else:
        raise Exception('Illegal engine {}, it should be '
                        'classifier or detector'.format(args.engine))

    for i in range(5):
        bgr_array = cv2.imread(args.input)
        t = time.time()
        image_data = engine.process_input(bgr_array)
        output = engine.inference(image_data)
        model_outputs = engine.process_output(output)
        # Reference result
        #     input:
        #         darknet/data/dog.jpg
        #     output:
        #         Inference takes 0.7533011436462402 s
        #         Inference takes 0.5741658210754395 s
        #         Inference takes 0.6120760440826416 s
        #         Inference takes 0.6191139221191406 s
        #         Inference takes 0.5809791088104248 s
        #         label: bicycle  conf: 0.9563907980918884  (139, 116) (567, 429)
        #         label: car  conf: 0.8757821917533875  (459, 80) (690, 172)
        #         label: dog  conf: 0.869189441204071  (131, 218) (314, 539)
        #         label: car  conf: 0.40003547072410583  (698, 122) (724, 152)
        logger.debug('Inference takes {} s'.format(time.time() - t))

    if args.engine == 'classifier':
        for r in model_outputs['annotations']:
            logger.debug('label: {0}  conf: {1}'.format(
                r['label'],
                r['confidence']
            ))
    elif args.engine == 'detector':
        for r in model_outputs['annotations']:
            logger.debug('label: {0}  conf: {1}  ({2}, {3}) ({4}, {5})'.format(
                r['label'],
                r['confidence'],
                r['left'],
                r['top'],
                r['right'],
                r['bottom']
            ))
    else:
        raise Exception('Can not get result '
                        'from illegal engine {}'.format(args.engine))


if __name__ == '__main__':
    main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Pipeline service restarter.

Restart pipeline service after received config update notification.
"""

import argparse
import logging
import subprocess

from berrynet import logger
from berrynet.comm import Communicator


class PipelineRestarterService(object):
    def __init__(self, service_name, comm_config):
        self.service_name = service_name
        self.comm_config = comm_config
        self.comm_config['subscribe']['dlboxapi/config/update'] = \
            self.restart_pipeline
        self.comm = Communicator(self.comm_config, debug=True)

    def restart_pipeline(self, pl):
        logger.debug('Restart pipeline')
        subprocess.call('dlbox-manager restart dlbox-pipeline.service',
                        shell=True)

    def run(self, args):
        """Infinite loop serving pipeline restart requests"""
        self.comm.run()


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--debug',
                    action='store_true',
                    help='Debug mode toggle')
    return vars(ap.parse_args())


def main():
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    comm_config = {
        'subscribe': {},
        'broker': {
            'address': 'localhost',
            'port': 1883
        }
    }
    engine_service = PipelineRestarterService(
        'pipeline service restarter',
        comm_config)
    engine_service.run(args)


if __name__ == '__main__':
    main()



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Simple image classification server with Inception.

The server monitors image_dir and run inferences on new images added to the
directory. Every image file should come with another empty file with '.done'
suffix to signal readiness. Inference result of a image can be read from the
'.txt' file of that image after '.txt.done' is spotted.

This is an example the server expects clients to do. Note the order.

# cp cat.jpg /run/image_dir
# touch /run/image_dir/cat.jpg.done

Clients should wait for appearance of 'cat.jpg.txt.done' before getting
result from 'cat.jpg.txt'.
"""


from __future__ import print_function

import argparse
import multiprocessing
import os
import signal
import sys
import threading
import time

from datetime import datetime

import cv2
import movidius as mv
import numpy as np
import Queue

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler


image_queue = Queue.Queue()
threads = []
mvng = None
SystemSnapshot = '/usr/local/berrynet/dashboard/www/freeboard/snapshot.jpg'


def logging(*args):
    print("[%08.3f]" % time.time(), ' '.join(args))


def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)


def server():
    """Infinite loop serving inference requests"""

    global image_queue
    global mvng

    logging(threading.current_thread().getName(), "is running")

    while True:
        input_name = image_queue.get()
        image_data = cv2.imread(input_name)
        processed_data = mv.process_yolo_input(image_data)

        t_start = datetime.now()
        output = mvng.inference(processed_data)
        t_end = datetime.now()
        t_inference = t_end - t_start
        logging('inference time: {} ms'.format(t_inference.total_seconds() * 1000))
        interpreted_output = mv.interpret_yolo_output(output,
                                                      image_data.shape[1],
                                                      image_data.shape[0])
        yolo_outputs = mv.process_yolo_output(image_data,
                                              interpreted_output,
                                              SystemSnapshot)

        output_name = input_name + '.txt'
        output_done_name = output_name + '.done'
        mv.save_yolo_output_text(output_name, yolo_outputs)
        touch(output_done_name)
        logging(input_name, " detected!")


class EventHandler(PatternMatchingEventHandler):
    def process(self, event):
        """
        event.event_type
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        global image_queue

        _msg = event.src_path
        image_queue.put(_msg.rstrip('.done'))
        os.remove(_msg)
        logging(_msg, event.event_type)

    # ignore all other types of events except 'modified'
    def on_created(self, event):
        self.process(event)


def main(args):
    global threads

    # Create a server thread for each CPU core
    cpu_count = multiprocessing.cpu_count()
    for i in xrange(cpu_count/4):
        threads.append(
            threading.Thread(target=server,
                             name='Server thread %d' % i))
    for t in threads:
        t.start()
    for t in threads:
        t.join()


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--model', required=True, help='Model file path')
    ap.add_argument('--label', required=True, help='Label file path')
    ap.add_argument('--image_dir', required=True, help='Path to image file')
    ap.add_argument('--num_top_predictions', default=5,
                    help='Display this many predictions')
    return vars(ap.parse_args())


if __name__ == '__main__':
    args = parse_args()

    mvng = mv.MovidiusNeuralGraph(args['model'], args['label'])

    pid = str(os.getpid())
    pidfile = "/tmp/detect_movidius_server.pid"

    if os.path.isfile(pidfile):
        logging("%s already exists, exiting" % pidfile)
        sys.exit(1)

    with open(pidfile, 'w') as f:
        f.write(pid)

    logging("model filepath: ", args['model'])
    logging("label filepath: ", args['label'])
    logging("image_dir: ", args['image_dir'])

    # workaround the issue that SIGINT cannot be received (fork a child to
    # avoid blocking the main process in Thread.join()
    child_pid = os.fork()
    if child_pid == 0:  # child
        # observer handles event in a different thread
        observer = Observer()
        observer.schedule(EventHandler(['*.jpg.done']), path=args['image_dir'])
        observer.start()
        main(args)
    else:  # parent
        try:
            os.wait()
        except KeyboardInterrupt:
            os.kill(child_pid, signal.SIGKILL)
            os.unlink(pidfile)



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Simple image classification server with Inception.

The server monitors image_dir and run inferences on new images added to the
directory. Every image file should come with another empty file with '.done'
suffix to signal readiness. Inference result of a image can be read from the
'.txt' file of that image after '.txt.done' is spotted.

This is an example the server expects clients to do. Note the order.

# cp cat.jpg /run/image_dir
# touch /run/image_dir/cat.jpg.done

Clients should wait for appearance of 'cat.jpg.txt.done' before getting
result from 'cat.jpg.txt'.
"""


from __future__ import print_function
import os
import sys
import time
import numpy as np
import threading
import multiprocessing
import tensorflow as tf
import Queue
import signal
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

FLAGS = tf.app.flags.FLAGS
image_queue = Queue.Queue()
sess = None
threads = []

# classify_image_graph_def.pb:
#   Binary representation of the GraphDef protocol buffer.
# imagenet_synset_to_human_label_map.txt:
#   Map from synset ID to a human readable string.
# imagenet_2012_challenge_label_map_proto.pbtxt:
#   Text representation of a protocol buffer mapping a label to synset ID.
tf.app.flags.DEFINE_string(
    'model_dir', 'model',
    """Path to output_graph.pb and output_labels.txt.""")
tf.app.flags.DEFINE_string('image_dir', 'image',
                           """Path to image file.""")
tf.app.flags.DEFINE_string('output_layer', 'softmax:0',
                           """Name of the result operation""")
tf.app.flags.DEFINE_string('input_layer', 'DecodeJpeg/contents:0',
                           """Name of the input operation""")
tf.app.flags.DEFINE_integer('num_top_predictions', 5,
                            """Display this many predictions.""")


def logging(*args):
    print("[%08.3f]" % time.time(), ' '.join(args))

def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)

def load_labels(filename):
    """Read in labels, one label per line."""
    return [line.rstrip() for line in tf.gfile.FastGFile(filename)]


def create_graph():
    """Creates a graph from saved GraphDef file and returns a saver."""
    # Creates graph from saved graph_def.pb.
    with tf.gfile.FastGFile(os.path.join(
        FLAGS.model_dir, 'output_graph.pb'), 'rb') as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())
        _ = tf.import_graph_def(graph_def, name='')


def server(labels):
    """Infinite loop serving inference requests"""

    global image_queue, sess

    logging(threading.current_thread().getName(), "is running")

    with sess.as_default():
        # Some useful tensors:
        # 'softmax:0': A tensor containing the normalized prediction across
        #   1000 labels.
        # 'pool_3:0': A tensor containing the next-to-last layer containing 2048
        #   float description of the image.
        # 'DecodeJpeg/contents:0': A tensor containing a string providing JPEG
        #   encoding of the image.

        while True:
            input_name = image_queue.get()
            image_data = tf.gfile.FastGFile(input_name, 'rb').read()

            predictions = sess.run(FLAGS.output_layer,
                                   {FLAGS.input_layer: image_data})
            predictions = np.squeeze(predictions)
            top_k = predictions.argsort()[-FLAGS.num_top_predictions:][::-1]

            output_name = input_name+'.txt'
            output_done_name = output_name+'.done'
            output = open(output_name, 'w')
            for node_id in top_k:
                human_string = labels[node_id]
                score = predictions[node_id]
                print("%s (score = %.5f)" % (human_string, score), file=output)
            output.close()
            touch(output_done_name)
            logging(input_name, " classified!")


class EventHandler(PatternMatchingEventHandler):
    def process(self, event):
        """
        event.event_type
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        global image_queue

        _msg = event.src_path
        image_queue.put(_msg.rstrip('.done'))
        os.remove(_msg)
        logging(_msg, event.event_type)

    # ignore all other types of events except 'modified'
    def on_created(self, event):
        self.process(event)


def main(_):
    """Called by Tensorflow"""

    global sess, threads

    # Creates graph from saved GraphDef.
    create_graph()

    # Reuse the same session for all threads processing requests
    sess = tf.Session()
    # Creates node ID --> English string lookup.
    labels = load_labels(os.path.join(FLAGS.model_dir, 'output_labels.txt'))

    # Create a server thread for each CPU core
    cpu_count = multiprocessing.cpu_count()
    for i in xrange(cpu_count/4):
        threads.append(threading.Thread(target=server,
                                        name='Server thread %d' % i,
                                        args=(labels,)))
    for t in threads: t.start()
    for t in threads: t.join()


if __name__ == '__main__':
    pid = str(os.getpid())
    pidfile = "/tmp/classify_server.pid"

    if os.path.isfile(pidfile):
        logging("%s already exists, exiting" % pidfile)
        sys.exit(1)

    with open(pidfile, 'w') as f:
        f.write(pid)

    logging("model_dir: ", FLAGS.model_dir)
    logging("image_dir: ", FLAGS.image_dir)

    # workaround the issue that SIGINT cannot be received (fork a child to
    # avoid blocking the main process in Thread.join()
    child_pid = os.fork()
    if child_pid == 0:
        # child
        # observer handles event in a different thread
        observer = Observer()
        observer.schedule(EventHandler(['*.jpg.done']), path=FLAGS.image_dir)
        observer.start()
        tf.app.run()
    else:
        # parent
        try:
            os.wait()
        except KeyboardInterrupt:
            os.kill(child_pid, signal.SIGKILL)
            os.unlink(pidfile)



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
# Copyright 2016 dt42.io. All Rights Reserved.
#
# 09-01-2016 joseph@dt42.io Initial version
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Simple image classification server with Inception.

The server monitors image_dir and run inferences on new images added to the
directory. Every image file should come with another empty file with '.done'
suffix to signal readiness. Inference result of a image can be read from the
'.txt' file of that image after '.txt.done' is spotted.

This is an example the server expects clients to do. Note the order.

# cp cat.jpg /run/image_dir
# touch /run/image_dir/cat.jpg.done

Clients should wait for appearance of 'cat.jpg.txt.done' before getting
result from 'cat.jpg.txt'.
"""


from __future__ import print_function
import os
import sys
import time
import signal
import argparse
import subprocess
import errno
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler


def logging(*args):
    print("[%08.3f]" % time.time(), ' '.join(args))


class EventHandler(PatternMatchingEventHandler):
    def process(self, event):
        """
        event.event_type
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        _msg = event.src_path
        os.remove(_msg)
        logging(_msg, event.event_type)
        darknet.stdin.write(_msg.rstrip('.done').encode('utf8') + b'\n')

    # ignore all other types of events except 'modified'
    def on_created(self, event):
        self.process(event)


def check_pid(pid):
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM means no permission, and the process exists to deny the
            # access
            return True
        else:
            raise
    else:
        return True

if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    pid = str(os.getpid())
    basename = os.path.splitext(os.path.basename(__file__))[0]
    ap.add_argument('input_dir')
    ap.add_argument(
        '-p', '--pid', default='/tmp/%s.pid' % basename,
        help='pid file path')
    ap.add_argument(
        '-fi', '--fifo', default='/tmp/acti_yolo',
        help='fifo pipe path')
    ap.add_argument(
        '-d', '--data', default='cfg/coco.data',
        help='fifo pipe path')
    ap.add_argument(
        '-c', '--config', default='cfg/yolo.cfg',
        help='fifo pipe path')
    ap.add_argument(
        '-w', '--weight', default='yolo.weights',
        help='fifo pipe path')
    args = vars(ap.parse_args())
    WATCH_DIR = os.path.abspath(args['input_dir'])
    FIFO_PIPE = os.path.abspath(args['fifo'])
    data = args['data']
    cfg = args['config']
    weight = args['weight']
    pidfile = args['pid']

    if os.path.isfile(pidfile):
        with open(pidfile) as f:
            prev_pid = int(f.readline())
            if check_pid(prev_pid):
                logging("{} already exists and process {} is still running, exists.".format(
                    pidfile, prev_pid))
                sys.exit(1)
            else:
                logging("{} exists but process {} died, clean it up.".format(pidfile, prev_pid))
                os.unlink(pidfile)

    with open(pidfile, 'w') as f:
        f.write(pid)

    logging("watch_dir: ", WATCH_DIR)
    logging("pid: ", pidfile)

    cmd = ['./darknet', 'detector', 'test', data, cfg, weight, '-out', '/usr/local/berrynet/dashboard/www/freeboard/snapshot']
    darknet = subprocess.Popen(cmd, bufsize=0,
            stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT)

    observer = Observer()
    observer.schedule(
        EventHandler(['*.jpg.done', '*.png.done']),
        path=WATCH_DIR, recursive=True)
    observer.start()
    try:
        darknet.wait()
    except KeyboardInterrupt:
        logging("Interrupt by user, clean up")
        os.kill(darknet.pid, signal.SIGKILL)
        os.unlink(pidfile)



#!/usr/bin/python3

import paho.mqtt.client as mqtt
import paho.mqtt.publish as publish

from berrynet import logger
from logzero import setup_logger


def on_connect(client, userdata, flags, rc):
    logger.debug('Connected with result code ' + str(rc))
    for topic in client.comm_config['subscribe'].keys():
        logger.debug('Subscribe topic {}'.format(topic))
        client.subscribe(topic)


def on_message(client, userdata, msg):
    """Dispatch received message to its bound functor.
    """
    logger.debug('Receive message from topic {}'.format(msg.topic))
    #logger.debug('Message payload {}'.format(msg.payload))
    client.comm_config['subscribe'][msg.topic](msg.payload)


class Communicator(object):
    def __init__(self, comm_config, debug=False):
        self.client = mqtt.Client()
        self.client.comm_config = comm_config
        self.client.on_connect = on_connect
        self.client.on_message = on_message

    def run(self):
        self.client.connect(
            self.client.comm_config['broker']['address'],
            self.client.comm_config['broker']['port'],
            60)
        self.client.loop_forever()

    def start_nb(self):
        self.client.connect(
            self.client.comm_config['broker']['address'],
            self.client.comm_config['broker']['port'],
            60)
        self.client.loop_start()

    def stop_nb(self):
        self.client.loop_stop()

    def send(self, topic, payload):
        logger.debug('Send message to topic {}'.format(topic))
        #logger.debug('Message payload {}'.format(payload))
        publish.single(topic, payload,
                       hostname=self.client.comm_config['broker']['address'])

    def disconnect(self):
        self.client.disconnect()



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""
DL Model Manager, following the DLModelBox model package
speccification.
"""

from __future__ import print_function

import argparse
import json
import os

from berrynet import logger


class DLModelManager(object):
    def __init__(self):
        self.basedir = '/usr/share/dlmodels'

    def get_model_names(self):
        return os.listdir(self.basedir)

    def get_model_meta(self, modelname):
        meta_filepath = os.path.join(self.basedir, modelname, 'meta.json')
        with open(meta_filepath, 'r') as f:
            meta = json.load(f)
        meta['model'] = os.path.join(self.basedir, modelname, meta['model'])
        meta['label'] = os.path.join(self.basedir, modelname, meta['label'])
        for k, v in meta['config'].items():
            meta['config'][k] = os.path.join(self.basedir, modelname,
                                             meta['config'][k])
        return meta


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--modelname',
                    help='Model package name (without version)')
    return vars(ap.parse_args())


if __name__ == '__main__':
    args = parse_args()
    logger.debug('model package name: ', args['modelname'])

    dlmm = DLModelManager()
    for name in dlmm.get_model_names():
        print(dlmm.get_model_meta(name))



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""
Snapshot service will listen to a trigger event (MQTT topic),
and send a snapshot retrieved from camera.
"""

import argparse
import json

from datetime import datetime

import cv2

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


class SnapshotService(object):
    def __init__(self, comm_config):
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = eval(functor)
        self.comm_config['subscribe']['berrynet/trigger/controller/snapshot'] = self.snapshot
        self.comm = Communicator(self.comm_config, debug=True)

    def snapshot(self, pl):
        '''Send camera snapshot.

        The functionality is the same as using camera client in file mode.

        The difference is that snapshot client retrieves image from camera
        instead of given filepath.
        '''
        duration = lambda t: (datetime.now() - t).microseconds / 1000

        # WORKAROUND: Prevent VideoCapture from buffering frames.
        #     VideoCapture will buffer frames automatically, and we need
        #     to find a way to disable it.
        self.capture = cv2.VideoCapture(0)
        status, im = self.capture.read()
        if (status is False):
            logger.warn('ERROR: Failure happened when reading frame')

        t = datetime.now()
        retval, jpg_bytes = cv2.imencode('.jpg', im)
        mqtt_payload = payload.serialize_jpg(jpg_bytes)
        self.comm.send('berrynet/data/rgbimage', mqtt_payload)
        logger.debug('send: {} ms'.format(duration(t)))
        self.capture.release()

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.run()


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument(
        '--topic-config',
        default=None,
        help='Path of the MQTT topic subscription JSON.'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()

    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        }
    }
    dc_service = SnapshotService(comm_config)
    dc_service.run(args)


if __name__ == '__main__':
    main()



#!/usr/bin/python

import json
import subprocess


def loadConfig():
    cmd = ("node -e \"const config = "
        "require('../config'); console.log(JSON.stringify(config));\"")
    config = json.loads(subprocess.check_output(cmd, shell=True))
    print(config)


if __name__ == '__main__':
    loadConfig()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""
Deep learning engine template provides unified interfaces for
different backends (e.g. TensorFlow, Caffe2, etc.)
"""

class DLEngine(object):
    def __init__(self):
        self.model_input_cache = []
        self.model_output_cache = []
        self.cache = {
            'model_input': [],
            'model_output': '',
            'model_output_filepath': ''
        }

    def create(self):
        # Workaround to posepone TensorFlow initialization.
        # If TF is initialized in __init__, and pass an engine instance
        # to engine service, TF session will stuck in run().
        pass

    def process_input(self, tensor):
        return tensor

    def inference(self, tensor):
        output = None
        return output

    def process_output(self, output):
        return output

    def cache_data(self, key, value):
        self.cache[key] = value

    def save_cache(self):
        with open(self.cache['model_output_filepath'], 'w') as f:
            f.write(str(self.cache['model_output']))



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Simple image detection server with Tiny-YOLO.

The server monitors image_dir and run inferences on new images added to the
directory. Every image file should come with another empty file with '.done'
suffix to signal readiness. Inference result of a image can be read from the
'.txt' file of that image after '.txt.done' is spotted.

This is an example the server expects clients to do. Note the order.

# cp cat.jpg /run/image_dir
# touch /run/image_dir/cat.jpg.done

Clients should wait for appearance of 'cat.jpg.txt.done' before getting
result from 'cat.jpg.txt'.
"""

from __future__ import print_function

import logging
import multiprocessing
import os
import Queue
import signal
import sys
import threading
import time

import cv2
import numpy as np

from os.path import join as pjoin
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from darkflow.net.build import TFNet


image_queue = Queue.Queue()

# FIXME: Make these variables configurable
ImageDir = '../image'
SystemSnapshot = '../../dashboard/www/freeboard/snapshot.jpg'


def _logging(*args):
    print("[%08.3f]" % time.time(), ' '.join(args))


def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)


def drawBoundingBoxes(imageData, imageOutputPath, inferenceResults, colorMap):
    """Draw bounding boxes on an image.

    imageData: image data in numpy array format
    imageOutputPath: output image file path
    inferenceResults: Darkflow inference results
    colorMap: Bounding box color candidates, list of RGB tuples.
    """
    # TODO: return raw data instead of save image
    for res in inferenceResults:
        left = res['topleft']['x']
        top = res['topleft']['y']
        right = res['bottomright']['x']
        bottom = res['bottomright']['y']
        colorIndex = res['coloridx']
        color = colorMap[colorIndex]
        label = res['label']
        confidence = res['confidence']
        imgHeight, imgWidth, _ = imageData.shape
        thick = int((imgHeight + imgWidth) // 300)

        cv2.rectangle(imageData,(left, top), (right, bottom), color, thick)
        cv2.putText(imageData, label, (left, top - 12), 0, 1e-3 * imgHeight,
            color, thick//3)
    cv2.imwrite(imageOutputPath, imageData)
    logging.debug('Save bounding box result image to {}'.format(imageOutputPath))


def server(tfnet):
    """Infinite loop serving inference requests"""

    global image_queue

    _logging(threading.current_thread().getName(), "is running")

    while True:
        input_name = image_queue.get()

        _logging('input image ' + input_name)
        imgcv = cv2.imread(input_name)
        _logging('start inference')
        result = tfnet.return_predict(imgcv)
        _logging('inference result: {}'.format(result))

        # overwrite existing input snapshot by the result image with
        # bounding boxes.
        drawBoundingBoxes(imgcv, SystemSnapshot, result, tfnet.meta['colors'])
        logging.debug('System snapshot path: %s' % pjoin(os.getcwd(), SystemSnapshot))

        output_name = input_name+'.txt'
        output_done_name = output_name+'.done'
        with open(output_name, 'w') as f:
            f.write(str(result))
        touch(output_done_name)
        _logging(input_name, " classified!")


class EventHandler(PatternMatchingEventHandler):
    def process(self, event):
        """
        event.event_type
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        # the file will be processed there
        global image_queue

        _msg = event.src_path
        image_queue.put(_msg.rstrip('.done'))
        os.remove(_msg)
        _logging(_msg, event.event_type)

    # ignore all other types of events except 'modified'
    def on_created(self, event):
        self.process(event)


def main():
    options = {
        "model": "cfg/tiny-yolo.cfg",
        "load": "bin/tiny-yolo.weights",
        'verbalise': True,
        #"threshold": 0.1
    }
    tfnet = TFNet(options)
    _logging('model dir: {}'.format(options['load']))
    _logging('config dir: {}'.format(options['model']))

    server(tfnet)


if __name__ == '__main__':
    logging.basicConfig(filename='/tmp/dlDetector.log', level=logging.DEBUG)

    pid = str(os.getpid())
    pidfile = "/tmp/detection_server.pid"

    if os.path.isfile(pidfile):
        _logging("%s already exists, exiting" % pidfile)
        sys.exit(1)

    with open(pidfile, 'w') as f:
        f.write(pid)

    # workaround the issue that SIGINT cannot be received (fork a child to
    # avoid blocking the main process in Thread.join()
    child_pid = os.fork()
    if child_pid == 0:
        # child
        # observer handles event in a different thread
        observer = Observer()
        observer.schedule(EventHandler(['*.jpg.done']), ImageDir)
        observer.start()
        main()
    else:
        # parent
        try:
            os.wait()
        except KeyboardInterrupt:
            os.kill(child_pid, signal.SIGKILL)
            os.unlink(pidfile)



#!/usr/bin/env python3
#
# Copyright 2019 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import io
import logging
import os
import tempfile
import tarfile
import time
import sys

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


class DydaConfigUpdateClient(object):
    def __init__(self, comm_config, debug=False):
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = self.handleResult
        self.comm = Communicator(self.comm_config, debug=True)

    def sendConfig(self, payloadID):
        self.comm.send(self.comm_config['publish'], payloadID)
        
    def handleResult(self, pl):
        try:
            payload_json = payload.deserialize_payload(pl.decode('utf-8'))
            print(payload_json)
            self.comm.stop_nb()
            sys.exit(0)
        except Exception as e:
            logger.info(e)

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.start_nb()
        self.sendConfig(args['payload'])
        time.sleep(1)

def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument('--debug',
        action='store_true',
        help='Debug mode toggle'
    )
    ap.add_argument(
        '--payload',
        required=True,
        help='payload ID'
    )
    ap.add_argument(
        '--topic',
        default='berrynet/manager/aikea/config',
        help='topic to listen for the result'
    )
    ap.add_argument(
        '--publish',
        default='berrynet/config/aikea/update',
        help='topic to publish'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    comm_config = {
        'subscribe': {
            args['topic']: None
        },
        'publish': args['publish'],
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        }
    }
    config_client = DydaConfigUpdateClient(comm_config,
                                        args['debug'])
    config_client.run(args)


if __name__ == '__main__':
    main()



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""TensorFlow inference engine.
"""

from __future__ import print_function

import argparse
import json

import cv2
import numpy as np
import tensorflow as tf

from berrynet import logger
#from berrynet.dlmodelmgr import DLModelManager
from berrynet.engine import DLEngine


class TensorFlowEngine(DLEngine):
    # FIXME: Get model information by model manager
    def __init__(self, model, label, input_layer, output_layer, top_k=3):
        super(TensorFlowEngine, self).__init__()

        # Load model
        with tf.gfile.FastGFile(model, 'rb') as f:
            graph_def = tf.GraphDef()
            graph_def.ParseFromString(f.read())
            _ = tf.import_graph_def(graph_def, name='')

        # Load labels
        self.labels = [line.rstrip() for line in tf.gfile.FastGFile(label)]

        # Load other configs
        self.input_layer = input_layer
        self.output_layer = output_layer
        self.top_k = top_k

        # NOTE: Do NOT call read_tensor_from_nparray twice to prevent from
        #       recreating unexpected placeholders.
        self.tensor_op = self.read_tensor_from_nparray(
            input_height=299,
            input_width=299,
            input_mean=0,
            input_std=255)

    def create(self):
        # Create session
        self.sess = tf.Session()

    def process_input(self, rgb_array):
        return self.sess.run(self.tensor_op,
                             feed_dict={'inarray:0': rgb_array})

    def inference(self, tensor):
        return self.sess.run(self.output_layer,
                             {self.input_layer: tensor})

    def process_output(self, output):
        processed_output = {'annotations': []}
        decimal_digits = 2
        predictions = np.squeeze(output)
        top_k_index = predictions.argsort()[-self.top_k:][::-1]

        for node_id in top_k_index:
            human_string = self.labels[node_id]
            score = round(float(predictions[node_id]), decimal_digits)
            anno = {
                'type': 'classification',
                'label': human_string,
                'confidence': score
            }
            processed_output['annotations'].append(anno)
            logger.debug('%s (score = %.5f)' % (human_string, score))
        return processed_output

    def save_cache(self):
        pass

    # NOTE: Copied from trainer.component
    def read_tensor_from_nparray(self, input_height=192, input_width=192,
                                 input_mean=0, input_std=255):
        """ Create normalized tensor based on input numpy array """
        image_reader = tf.placeholder(tf.uint8, name='inarray')
        float_caster = tf.cast(image_reader, tf.float32)
        dims_expander = tf.expand_dims(float_caster, 0)
        resized = tf.image.resize_bilinear(dims_expander,
                                           [input_height, input_width])
        normalized = tf.divide(tf.subtract(resized, [input_mean]), [input_std])
        return normalized



import os
import sys
import time

from datetime import datetime

import cv2
import numpy as np

from mvnc import mvncapi as mvnc
from skimage.transform import resize


def interpret_yolo_output(output, img_width, img_height):
    output = output.astype(np.float32)

    classes = [
        'aeroplane', 'bicycle', 'bird', 'boat', 'bottle',
        'bus', 'car', 'cat', 'chair', 'cow',
        'diningtable', 'dog', 'horse', 'motorbike', 'person',
        'pottedplant', 'sheep', 'sofa', 'train','tvmonitor'
    ]
    threshold = 0.2
    iou_threshold = 0.5
    num_class = 20
    num_box = 2
    grid_size = 7
    probs = np.zeros((7, 7, 2, 20))
    class_probs = (np.reshape(output[0:980], (7, 7, 20)))  #.copy()
    scales = (np.reshape(output[980:1078], (7, 7, 2)))  #.copy()
    boxes = (np.reshape(output[1078:], (7, 7, 2, 4)))  #.copy()
    offset = np.transpose(
        np.reshape(np.array([np.arange(7)] * 14), (2, 7, 7)),
        (1, 2, 0)
    )
    #boxes.setflags(write=1)
    boxes[:, :, :, 0] += offset
    boxes[:, :, :, 1] += np.transpose(offset, (1, 0, 2))
    boxes[:, :, :, 0:2] = boxes[:, :, :, 0:2] / 7.0
    boxes[:, :, :, 2] = np.multiply(boxes[:, :, :, 2], boxes[:, :, :, 2])
    boxes[:, :, :, 3] = np.multiply(boxes[:, :, :, 3], boxes[:, :, :, 3])

    boxes[:, :, :, 0] *= img_width
    boxes[:, :, :, 1] *= img_height
    boxes[:, :, :, 2] *= img_width
    boxes[:, :, :, 3] *= img_height

    for i in range(2):
        for j in range(20):
            probs[:, :, i, j] = np.multiply(class_probs[:, :, j],
                                            scales[:, :, i])
    #print (probs)
    filter_mat_probs = np.array(probs >= threshold, dtype='bool')
    filter_mat_boxes = np.nonzero(filter_mat_probs)
    boxes_filtered = boxes[filter_mat_boxes[0],
                           filter_mat_boxes[1],
                           filter_mat_boxes[2]]
    probs_filtered = probs[filter_mat_probs]
    classes_num_filtered = np.argmax(probs, axis=3)[filter_mat_boxes[0],
                                                    filter_mat_boxes[1],
                                                    filter_mat_boxes[2]]

    argsort = np.array(np.argsort(probs_filtered))[::-1]
    boxes_filtered = boxes_filtered[argsort]
    probs_filtered = probs_filtered[argsort]
    classes_num_filtered = classes_num_filtered[argsort]

    for i in range(len(boxes_filtered)):
        if probs_filtered[i] == 0:
            continue
        for j in range(i + 1, len(boxes_filtered)):
            if iou(boxes_filtered[i], boxes_filtered[j]) > iou_threshold:
                probs_filtered[j] = 0.0

    filter_iou = np.array(probs_filtered > 0.0, dtype='bool')
    boxes_filtered = boxes_filtered[filter_iou]
    probs_filtered = probs_filtered[filter_iou]
    classes_num_filtered = classes_num_filtered[filter_iou]

    result = []
    for i in range(len(boxes_filtered)):
        result.append([classes[classes_num_filtered[i]],
                       boxes_filtered[i][0],
                       boxes_filtered[i][1],
                       boxes_filtered[i][2],
                       boxes_filtered[i][3],
                       probs_filtered[i]])

    return result


def iou(box1, box2):
    tb = (min(box1[0] + 0.5 * box1[2], box2[0] + 0.5 * box2[2]) -
          max(box1[0] - 0.5 * box1[2], box2[0] - 0.5 * box2[2]))
    lr = (min(box1[1] + 0.5 * box1[3], box2[1] + 0.5 * box2[3]) -
          max(box1[1] - 0.5 * box1[3], box2[1] - 0.5 * box2[3]))
    if tb < 0 or lr < 0:
        intersection = 0
    else:
        intersection =  tb*lr
    return intersection / (box1[2] * box1[3] + box2[2] * box2[3] - intersection)


def process_yolo_output(img, results):
    img_width = img.shape[1]
    img_height = img.shape[0]
    img_cp = img.copy()

    print_yolo_output(results)

    # draw bounding boxes on input image
    for i in range(len(results)):
        x = int(results[i][1])
        y = int(results[i][2])
        w = int(results[i][3]) // 2
        h = int(results[i][4]) // 2
        xmin = x - w
        xmax = x + w
        ymin = y - h
        ymax = y + h
        if xmin < 0:
            xmin = 0
        if ymin < 0:
            ymin = 0
        if xmax > img_width:
            xmax = img_width
        if ymax > img_height:
            ymax = img_height
        cv2.rectangle(img_cp,
                      (xmin, ymin),
                      (xmax, ymax),
                      (0, 255, 0),
                      2)
        cv2.rectangle(img_cp,
                      (xmin, ymin - 20),
                      (xmax, ymin),
                      (125, 125, 125),
                      -1)
        cv2.putText(img_cp,results[i][0] + ' : %.2f' % results[i][5],
                    (xmin + 5, ymin - 7),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.5,
                    (0, 0, 0),
                    1)
    cv2.imwrite('/tmp/yolo_result.jpg', img_cp)


def process_yolo_input(rgb_data):
    input_dim = (448, 448)
    tmp_data = rgb_data.copy()
    tmp_data = resize(tmp_data / 255.0, input_dim, 1)
    tmp_data[:, :, (2, 1, 0)]  # BGR2RGB
    input_data = tmp_data.astype(np.float16)
    return input_data


def print_yolo_output(output):
    for i in range(len(output)):
        x = int(output[i][1])
        y = int(output[i][2])
        #w = int(output[i][3]) // 2
        #h = int(output[i][4]) // 2
        print('\tclass = {label}'.format(label=output[i][0]))
        print('\t[x, y, w, h] = [{x}, {y}, {w}, {h}]'.format(
            x=str(x),
            y=str(y),
            w=str(int(output[i][3])),
            h=str(int(output[i][4]))))
        print('\tconfidence = {conf}'.format(conf=str(results[i][5])))


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print ("YOLOv1 Tiny example: python3 py_examples/yolo_example.py images/dog.jpg")
        sys.exit()

    network_blob='/home/pi/codes/yoloNCS/graph'
    # configuration NCS
    mvnc.SetGlobalOption(mvnc.GlobalOption.LOG_LEVEL, 2)
    devices = mvnc.EnumerateDevices()
    if len(devices) == 0:
        print('No devices found')
        quit()
    device = mvnc.Device(devices[0])
    device.OpenDevice()
    opt = device.GetDeviceOption(mvnc.DeviceOption.OPTIMISATION_LIST)
    # load blob
    with open(network_blob, mode='rb') as f:
        blob = f.read()
    graph = device.AllocateGraph(blob)
    graph.SetGraphOption(mvnc.GraphOption.ITERATIONS, 1)
    iterations = graph.GetGraphOption(mvnc.GraphOption.ITERATIONS)

    # image preprocess
    img = cv2.imread(sys.argv[1])
    input_data = process_yolo_input(img)

    # start MOD
    start = datetime.now()
    graph.LoadTensor(input_data, 'user object')
    out, userobj = graph.GetResult()
    end = datetime.now()
    elapsedTime = end-start
    print('total time is " milliseconds', elapsedTime.total_seconds()*1000)

    # fc27 instead of fc12 for yolo_small
    results = interpret_yolo_output(out,
                                    img.shape[1],
                                    img.shape[0])
    #print (results)
    #cv2.imshow('YOLO detection',img_cv)
    process_yolo_output(img, results)
    #cv2.waitKey(10000)

    graph.DeallocateGraph()
    device.CloseDevice()



# Copyright 2019 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

# Reference
# https://www.geeksforgeeks.org/send-mail-attachment-gmail-account-using-python/

"""Gmail client sends an email with inference result.

The email will contain two attachments: image and text.
"""

import argparse
import json
import logging
import os
import smtplib

from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email import encoders

from os.path import join as pjoin

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


def create_mime_attachment(filepath):
    filename = os.path.basename(filepath)
    attachment = open(filepath, "rb")

    # instance of MIMEBase and named as p
    p = MIMEBase('application', 'octet-stream')
    # To change the payload into encoded form
    p.set_payload(attachment.read())
    # encode into base64
    encoders.encode_base64(p)
    p.add_header('Content-Disposition',
                 "attachment; filename= %s" % filename)
    return p


def send_email_text(sender_address,
                    sender_password,
                    receiver_address,
                    body='',
                    subject='BerryNet mail client notification',
                    attachments=None):
    # instance of MIMEMultipart
    msg = MIMEMultipart()

    msg['From'] = sender_address
    msg['To'] = receiver_address
    msg['Subject'] = subject
    logger.debug('Sender: {}'.format(msg['From']))
    logger.debug('Receiver: {}'.format(msg['To']))
    logger.debug('Subject: {}'.format(msg['Subject']))

    # attach the body with the msg instance
    msg.attach(MIMEText(body, 'plain'))

    for fpath in attachments:
        logger.debug('Attachment: {}'.format(fpath))
        msg.attach(create_mime_attachment(fpath))

    # creates SMTP session
    s = smtplib.SMTP('smtp.gmail.com', 587)
    # start TLS for security
    s.starttls()
    # Authentication
    s.login(sender_address, sender_password)
    # Converts the Multipart msg into a string
    text = msg.as_string()
    # sending the mail
    s.sendmail(sender_address, receiver_address, text)
    # terminating the session
    s.quit()


class GmailService(object):
    def __init__(self, comm_config):
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = eval(functor)
        self.comm = Communicator(self.comm_config, debug=True)
        self.email = comm_config['email']
        self.pipeline_compatible = comm_config['pipeline_compatible']
        self.target_label = comm_config['target_label']

    def find_target_label(self, target_label, generalized_result):
        label_list = [i['label'] for i in generalized_result['annotations']]
        logger.debug('Result labels: {}'.format(label_list))
        return target_label in label_list

    def update(self, pl):
        payload_json = payload.deserialize_payload(pl.decode('utf-8'))
        if self.pipeline_compatible:
            b64img_key = 'image_blob'
        else:
            b64img_key = 'bytes'
        jpg_bytes = payload.destringify_jpg(payload_json[b64img_key])
        payload_json.pop(b64img_key)
        logger.debug('inference text result: {}'.format(payload_json))

        match_target_label = self.find_target_label(self.target_label,
                                                    payload_json)

        logger.debug('Find target label {0}: {1}'.format(
            self.target_label, match_target_label))

        if match_target_label:
            timestamp = datetime.now().isoformat()
            notification_image = pjoin('/tmp', timestamp + '.jpg')
            notification_text = pjoin('/tmp', timestamp + '.json')
            with open(notification_image, 'wb') as f:
                f.write(jpg_bytes)
            with open(notification_text, 'w') as f:
                f.write(json.dumps(payload_json, indent=4))

            try:
                send_email_text(
                    self.email['sender_address'],
                    self.email['sender_password'],
                    self.email['receiver_address'],
                    body=('Target label {} is found. '
                          'Please check the attachments.'
                          ''.format(self.target_label)),
                    subject='BerryNet mail client notification',
                    attachments=set([notification_image, notification_text]))
            except Exception as e:
                logger.warn(e)

            os.remove(notification_image)
            os.remove(notification_text)
        else:
            # target label is not in generalized result, do nothing
            pass

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.run()


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--sender-address',
        required=True,
        help='Email address of sender. Ex: foo@email.org'
    )
    ap.add_argument(
        '--sender-password',
        required=True,
        help='Password of sender email address.'
    )
    ap.add_argument(
        '--receiver-address',
        required=True,
        help='Email address of receiver. Ex: bar@email.org'
    )
    ap.add_argument(
        '--target-label',
        required=True,
        help='Send notification email if the label is in inference result.'
    )
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument(
        '--topic',
        nargs='*',
        default=['berrynet/engine/tflitedetector/result'],
        help='The topic to listen, and can be indicated multiple times.'
    )
    ap.add_argument(
        '--topic-action',
        default='self.update',
        help='The action for the indicated topics.'
    )
    ap.add_argument(
        '--topic-config',
        default=None,
        help='Path of the MQTT topic subscription JSON.'
    )
    ap.add_argument(
        '--pipeline-compatible',
        action='store_true',
        help=(
            'Change key of b64 image string in generalized result '
            'from bytes to image_blob. '
            'Note: This is an experimental parameter.'
        )
    )
    ap.add_argument('--debug',
        action='store_true',
        help='Debug mode toggle'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # Topics and actions can come from two sources: CLI and config file.
    # Setup topic_config by parsing values from the two sources.
    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    topic_config.update({t:args['topic_action'] for t in args['topic']})

    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        },
        'email': {
            'sender_address': args['sender_address'],
            'sender_password': args['sender_password'],
            'receiver_address': args['receiver_address']
        },
        'pipeline_compatible': args['pipeline_compatible'],
        'target_label': args['target_label']
    }
    dc_service = GmailService(comm_config)
    dc_service.run(args)


if __name__ == '__main__':
    main()





# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""OpenVINO classification inference engine.
"""

import logging
import os
import sys

from argparse import ArgumentParser
from time import time

from berrynet.engine import DLEngine
import cv2
import numpy as np

from berrynet import logger
from openvino.inference_engine import IENetwork, IEPlugin


class OpenVINOClassifierEngine(DLEngine):
    def __init__(self, model, labels=None, top_k=3, device='CPU'):
        """
        Args:
            model: Path to an .xml file with a trained model.

            device: Specify the target device to infer on; CPU, GPU, FPGA
                    or MYRIAD is acceptable. Sample will look for a suitable
                    plugin for device specified (CPU by default)

            labels: Labels mapping file

            top_k: Number of top results
        """
        super(OpenVINOClassifierEngine, self).__init__()

        model_bin = model
        model_xml = os.path.splitext(model_bin)[0] + ".xml"
        if labels:
            with open(labels, 'r') as f:
                # Allow label name with spaces. To use onlyh the 1st word,
                # uncomment another labels_map implementation below.
                self.labels_map = [l.strip() for l in f.readlines()]
                #self.labels_map = [x.split(sep=' ', maxsplit=1)[-1].strip()
                #                   for x in f]
        else:
            self.labels_map = None
        self.top_k = top_k

        # Plugin initialization for specified device and
        # load extensions library if specified
        #
        # Note: MKLDNN CPU-targeted custom layer support is not included
        #       because we do not use it yet.
        self.plugin = IEPlugin(device=device, plugin_dirs=None)

        # Read IR
        logger.debug('Loading network files:'
                     '\n\txml: {0}\n\tbin: {1}'.format(model_xml, model_bin))
        net = IENetwork.from_ir(model=model_xml, weights=model_bin)

        if self.plugin.device == "CPU":
            supported_layers = self.plugin.get_supported_layers(net)
            not_supported_layers = [l for l in net.layers.keys() if l not in supported_layers]
            if len(not_supported_layers) != 0:
                logger.error("Following layers are not supported by the plugin for specified device {}:\n {}".format(self.plugin.device, ', '.join(not_supported_layers)))
                sys.exit(1)

        assert len(net.inputs.keys()) == 1, "Sample supports only single input topologies"
        assert len(net.outputs) == 1, "Sample supports only single output topologies"

        # input_blob and and out_blob are the layer names in string format.
        logger.debug("Preparing input blobs")
        self.input_blob = next(iter(net.inputs))
        self.out_blob = next(iter(net.outputs))
        net.batch_size = 1

        self.n, self.c, self.h, self.w = net.inputs[self.input_blob].shape

        # Loading model to the plugin
        logger.debug("Loading model to the plugin")
        self.exec_net = self.plugin.load(network=net)

        del net

    def __delete__(self, instance):
        del self.exec_net
        del self.plugin

    def process_input(self, tensor):
        """Resize tensor (if needed) and change layout from HWC to CHW.

        Args:
            tensor: Input BGR tensor (OpenCV convention)

        Returns:
            Resized and transposed tensor
        """
        if tensor.shape[:-1] != (self.h, self.w):
            logger.warning("Input tensor is resized from {} to {}".format(
                tensor.shape[:-1], (self.h, self.w)))
            tensor = cv2.resize(tensor, (self.w, self.h))
        tensor = tensor.transpose((2, 0, 1))  # Change data layout from HWC to CHW
        return tensor

    def inference(self, tensor):
        logger.debug("Starting inference")
        res = self.exec_net.infer(inputs={self.input_blob: tensor})
        return res[self.out_blob]

    def process_output(self, output):
        logger.debug("Processing output blob")
        logger.debug("Top {} results: ".format(self.top_k))

        annotations = []
        for i, probs in enumerate(output):
            probs = np.squeeze(probs)
            top_ind = np.argsort(probs)[-self.top_k:][::-1]
            for id in top_ind:
                det_label = self.labels_map[id] if self.labels_map else "#{}".format(id)
                logger.debug("\t{:.7f} label {}".format(probs[id], det_label))

                annotations.append({
                    'type': 'classification',
                    'label': det_label,
                    'confidence': float(probs[id])
                })
        return {'annotations': annotations}


class OpenVINODetectorEngine(DLEngine):
    def __init__(self, model, labels=None, threshold=0.3, device='CPU'):
        super(OpenVINODetectorEngine, self).__init__()

        # Prepare model and labels
        model_bin = model
        model_xml = os.path.splitext(model_bin)[0] + ".xml"
        if labels:
            with open(labels, 'r') as f:
                self.labels_map = [x.strip() for x in f]
        else:
            self.labels_map = None
        self.threshold = threshold

        # Plugin initialization for specified device and
        # load extensions library if specified
        #
        # Note: MKLDNN CPU-targeted custom layer support is not included
        #       because we do not use it yet.
        if device == 'CPU':
            # Since OpenVINO 2020.1, cpu_extension content has been merged
            # into MKLDNN plugin and it is used automatically.
            #
            # We do not need to call these functions below anymore:
            #     self.plugin.add_cpu_extension(plugin_dirs + '/libcpu_extension_avx2.so')
            #     self.plugin.add_cpu_extension(plugin_dirs + '/libcpu_extension_avx512.so')
            #     self.plugin.add_cpu_extension(plugin_dirs + '/libcpu_extension_sse4.so')
            plugin_dirs = '/opt/intel/openvino/deployment_tools/inference_engine/lib/intel64'
            self.plugin = IEPlugin(device=device, plugin_dirs=plugin_dirs)
            logger.debug('plugin dirs: {}'.format(plugin_dirs))
        else:
            plugin_dirs = None
            self.plugin = IEPlugin(device=device, plugin_dirs=plugin_dirs)

        # Read IR
        logger.debug('Loading network files:'
                     '\n\txml: {0}\n\tbin: {1}'.format(model_xml, model_bin))
        net = IENetwork(model=model_xml, weights=model_bin)

        if self.plugin.device == "CPU":
            supported_layers = self.plugin.get_supported_layers(net)
            not_supported_layers = [l for l in net.layers.keys() if l not in supported_layers]
            if len(not_supported_layers) != 0:
                logger.error("Following layers are not supported by the plugin for specified device {}:\n {}".
                          format(self.plugin.device, ', '.join(not_supported_layers)))
                logger.error("Please try to specify cpu extensions library path in demo's command line parameters using -l "
                          "or --cpu_extension command line argument")
                sys.exit(1)

        assert len(net.inputs.keys()) == 1, "Demo supports only single input topologies"
        assert len(net.outputs) == 1, "Demo supports only single output topologies"

        # input_blob and and out_blob are the layer names in string format.
        logger.debug("Preparing input blobs")
        self.input_blob = next(iter(net.inputs))
        self.out_blob = next(iter(net.outputs))

        self.n, self.c, self.h, self.w = net.inputs[self.input_blob].shape

        # Loading model to the plugin
        self.exec_net = self.plugin.load(network=net, num_requests=2)

        del net

        # Initialize engine mode: sync or async
        #
        # FIXME: async mode does not work currently.
        #        process_input needs to provide two input tensors for async.
        self.is_async_mode = False
        self.cur_request_id = 0
        self.next_request_id = 1

    def __delete__(self, instance):
        del self.exec_net
        del self.plugin

    def process_input(self, tensor, next_tensor=None):
        frame = tensor
        next_frame = next_tensor

        # original input shape will be used in process_output
        self.img_w = tensor.shape[1]
        self.img_h = tensor.shape[0]

        # Main sync point:
        # in the truly Async mode we start the NEXT infer request, while waiting for the CURRENT to complete
        # in the regular mode we start the CURRENT request and immediately wait for it's completion
        if self.is_async_mode:
            in_frame = cv2.resize(next_frame, (self.w, self.h))
            in_frame = in_frame.transpose((2, 0, 1))  # Change data layout from HWC to CHW
            in_frame = in_frame.reshape((self.n, self.c, self.h, self.w))
        else:
            in_frame = cv2.resize(frame, (self.w, self.h))
            in_frame = in_frame.transpose((2, 0, 1))  # Change data layout from HWC to CHW
            in_frame = in_frame.reshape((self.n, self.c, self.h, self.w))
        return in_frame

    def inference(self, tensor):
        inf_start = time()
        if self.is_async_mode:
            self.exec_net.start_async(request_id=self.next_request_id, inputs={self.input_blob: tensor})
        else:
            self.exec_net.start_async(request_id=self.cur_request_id, inputs={self.input_blob: tensor})

        if self.exec_net.requests[self.cur_request_id].wait(-1) == 0:
            inf_end = time()
            det_time = inf_end - inf_start
            if self.is_async_mode:
                logger.debug('Inference time: N\A for async mode')
            else:
                logger.debug("Inference time: {:.3f} ms".format(det_time * 1000))

            # Parse detection results of the current request
            res = self.exec_net.requests[self.cur_request_id].outputs[self.out_blob]
        else:
            res = None

        return res

        # FIXME: async mode does not work currently.
        #        process_input needs to provide two input tensors for async.
        if self.is_async_mode:
            self.cur_request_id, self.next_request_id = self.next_request_id, self.cur_request_id
            frame = next_frame

    def process_output(self, output):
        logger.debug("Processing output blob")
        logger.debug("Threshold: {}".format(self.threshold))

        annotations = []
        for obj in output[0][0]:
            # Collect objects when probability more than specified threshold
            if obj[2] > self.threshold:
                xmin = int(obj[3] * self.img_w)
                ymin = int(obj[4] * self.img_h)
                xmax = int(obj[5] * self.img_w)
                ymax = int(obj[6] * self.img_h)
                class_id = int(obj[1])
                det_label = self.labels_map[class_id] if self.labels_map else str(class_id)
                annotations.append({
                    'label': det_label,
                    'confidence': float(obj[2]),
                    'left': xmin,
                    'top': ymin,
                    'right': xmax,
                    'bottom': ymax
                })
        return {'annotations': annotations}


def get_distribution_info():
    """Get Debuan or Ubuntu distribution information.
    """
    info = {}
    with open('/etc/lsb-release') as f:
        info_l = [i.strip().split('=') for i in f.readlines()]
    for i in info_l:
        info[i[0]] = i[1]
    logger.debug('Distribution info: {}'.format(info))
    return info


def set_openvino_environment():
    """The same effect as executing <openvino>/bin/setupvars.sh.
    """
    dist_info = get_distribution_info()
    python_version = 3.5

    os.environ['INSTALLDIR'] = '/opt/intel/computer_vision_sdk_2018.4.420'
    os.environ['INTEL_CVSDK_DIR'] = os.environ['INSTALLDIR']
    os.environ['LD_LIBRARY_PATH'] = (
        '{installdir}/deployment_tools/model_optimizer/bin:'
        '{ld_library_path}').format(
            installdir = os.environ.get('INSTALLDIR'),
            ld_library_path = os.environ.get('LD_LIBRARY_PATH' or '')
    )
    os.environ['InferenceEngine_DIR'] = os.path.join(
        os.environ.get('INTEL_CVSDK_DIR'),
        'deployment_tools/inference_engine/share'
    )
    os.environ['IE_PLUGINS_PATH'] = os.path.join(
        os.environ.get('INTEL_CVSDK_DIR'),
        'deployment_tools/inference_engine/lib/ubuntu_{}.04/intel64'.format(
            dist_info['DISTRIB_RELEASE'])
    )
    os.environ['LD_LIBRARY_PATH'] = (
        '/opt/intel/opencl:'
        '{installdir}/deployment_tools/inference_engine/external/gna/lib:'
        '{installdir}/deployment_tools/inference_engine/external/mkltiny_lnx/lib:'
        '{installdir}/deployment_tools/inference_engine/external/omp/lib:'
        '{ie_plugins_path}:'
        '{ld_library_path}').format(
            installdir = os.environ.get('INSTALLDIR'),
            ie_plugins_path = os.environ.get('IE_PLUGINS_PATH'),
            ld_library_path = os.environ.get('LD_LIBRARY_PATH' or '')
    )
    os.environ['PATH'] = (
        '{intel_cvsdk_dir}/deployment_tools/model_optimizer:'
        '{path}').format(
            intel_cvsdk_dir = os.environ.get('INTEL_CVSDK_DIR'),
            path = os.environ.get('PATH'),
    )
    os.environ['PYTHONPATH'] = (
        '{intel_cvsdk_dir}/deployment_tools/model_optimizer:'
        '{pythonpath}').format(
            intel_cvsdk_dir = os.environ.get('INTEL_CVSDK_DIR'),
            pythonpath = os.environ.get('PYTHONPATH' or '')
    )
    os.environ['PYTHONPATH'] = (
        '{intel_cvsdk_dir}/python/python$python_version:'
        '{intel_cvsdk_dir}/python/python$python_version/ubuntu16:'
        '{pythonpath}').format(
            intel_cvsdk_dir = os.environ.get('INTEL_CVSDK_DIR'),
            pythonpath = os.environ.get('PYTHONPATH' or '')
    )


def parse_argsr():
    parser = ArgumentParser()
    parser.add_argument(
        "-e", "--engine",
        help=("Classifier or Detector engine. "
              "classifier, or detector is acceptable. "
              "(classifier by default)"),
        default="classifier",
        type=str)
    parser.add_argument(
        "-m", "--model",
        help="Path to an .xml file with a trained model.",
        required=True,
        type=str)
    parser.add_argument(
        "-l", "--labels",
        help="Labels mapping file",
        default=None,
        type=str)
    parser.add_argument(
        "--top_k",
        help="Number of top results",
        default=10,
        type=int)
    parser.add_argument(
        "-d", "--device",
        help="Specify the target device to infer on; CPU, GPU, FPGA or MYRIAD is acceptable. Sample will look for a suitable plugin for device specified (CPU by default)",
        default="CPU",
        type=str)
    parser.add_argument(
        "-i", "--input",
        help="Path to a folder with images or path to an image files",
        required=True,
        type=str)
    parser.add_argument(
        "--debug",
        help="Debug mode toggle",
        default=False,
        action="store_true")

    return parser.parse_args()


def main():
    args = parse_argsr()

    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if args.engine == 'classifier':
        engine = OpenVINOClassifierEngine(
                     model = args.model,
                     device = args.device,
                     labels = args.labels,
                     top_k = args.top_k)
    elif args.engine == 'detector':
        engine = OpenVINODetectorEngine(
                     model = args.model,
                     device = args.device,
                     labels = args.labels)
    else:
        raise Exception('Illegal engine {}, it should be '
                        'classifier or detector'.format(args.engine))

    #set_openvino_environment()
    #if args.debug:
    #    logger.debug('OpenVINO environment vars')
    #    target_vars = ['INSTALLDIR',
    #                   'INTEL_CVSDK_DIR',
    #                   'LD_LIBRARY_PATH',
    #                   'InferenceEngine_DIR',
    #                   'IE_PLUGINS_PATH',
    #                   'PATH',
    #                   'PYTHONPATH']
    #    for i in target_vars:
    #        logger.debug('\t{var}: {val}'.format(
    #            var = i,
    #            val = os.environ.get(i)))

    bgr_array = cv2.imread(args.input)
    t = time()
    image_data = engine.process_input(bgr_array)
    output = engine.inference(image_data)
    model_outputs = engine.process_output(output)
    logger.debug('Result: {}'.format(model_outputs))
    logger.debug('Classification takes {} s'.format(time() - t))


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Data collector with UI showing inference result for human
"""

import argparse
import json
import os
import sys
import threading
import tkinter as tk

from datetime import datetime
from os.path import join as pjoin

import cv2
import numpy as np

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload
from PIL import Image
from PIL import ImageTk


class DataCollectorService(object):
    def __init__(self, comm_config, data_dirpath):
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = eval(functor)
        #self.comm_config['subscribe']['berrynet/data/rgbimage'] = self.update
        self.comm_config['subscribe']['berrynet/engine/pipeline/result'] = self.save_pipeline_result
        self.comm = Communicator(self.comm_config, debug=True)
        self.data_dirpath = data_dirpath

    def update(self, pl):
        payload_json = payload.deserialize_payload(pl.decode('utf-8'))

        # update UI with the latest inference result
        self.ui.update(payload_json, 'bytes')

        if self.data_dirpath:
            if not os.path.exists(self.data_dirpath):
                try:
                    os.mkdir(self.data_dirpath)
                except Exception as e:
                    logger.warn('Failed to create {}'.format(self.data_dirpath))
                    raise(e)

            jpg_bytes = payload.destringify_jpg(payload_json['bytes'])
            payload_json.pop('bytes')
            logger.debug('inference text result: {}'.format(payload_json))

            timestamp = datetime.now().isoformat()
            with open(pjoin(self.data_dirpath, timestamp + '.jpg'), 'wb') as f:
                f.write(jpg_bytes)
            with open(pjoin(self.data_dirpath, timestamp + '.json'), 'w') as f:
                f.write(json.dumps(payload_json, indent=4))

    def save_pipeline_result(self, pl):
        payload_json = payload.deserialize_payload(pl.decode('utf-8'))

        # update UI with the latest inference result
        self.ui.update(payload_json, 'image_blob')

        if self.data_dirpath:
            if not os.path.exists(self.data_dirpath):
                try:
                    os.mkdir(self.data_dirpath)
                except Exception as e:
                    logger.warn('Failed to create {}'.format(self.data_dirpath))
                    raise(e)

            jpg_bytes = payload.destringify_jpg(payload_json['image_blob'])
            payload_json.pop('image_blob')
            logger.debug('inference text result: {}'.format(payload_json))

            timestamp = datetime.now().isoformat()
            with open(pjoin(self.data_dirpath, timestamp + '.jpg'), 'wb') as f:
                f.write(jpg_bytes)
            with open(pjoin(self.data_dirpath, timestamp + '.json'), 'w') as f:
                f.write(json.dumps(payload_json, indent=4))

    def send_snapshot_trigger(self):
        payload = {}
        payload['timestamp'] = datetime.now().isoformat()
        mqtt_payload = json.dumps(payload)
        self.comm.send('berrynet/trigger/controller/snapshot', mqtt_payload)

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.run()


class UI(object):
    def __init__(self, dc_service, dc_kwargs):
        # Create data collector attributes
        self.dc_service = dc_service
        self.dc_kwargs = dc_kwargs
        self.dc_service.ui = self

        # Create UI attributes
        self.window = tk.Tk()
        self.window.title('BerryNet Inference Dashboard')
        self.window.protocol('WM_DELETE_WINDOW', self.on_closing)
        self.canvas_w = dc_kwargs['image_width']
        self.canvas_h = dc_kwargs['image_height']
        self.crowd_factor = 3

        # Add label: inference result text
        self.result = tk.Label(self.window,
                               text='TBD',
                               font=('Courier New', 10),
                               justify=tk.LEFT)
        #self.result.pack(expand=True, side=tk.LEFT)
        self.result.grid(row=0, column=0, padx=10)
        #self.result.columnconfigure(1, weight=2)

        # Add canvas: inference result image
        #self.canvas = tk.Canvas(self.window, width=1920, height=1080)
        self.canvas = tk.Canvas(self.window)
        self.photo = ImageTk.PhotoImage(
                         image=Image.fromarray(
                             np.zeros((self.canvas_h, self.canvas_w, 3), dtype=np.uint8)))
        self.image_id = self.canvas.create_image(
                            0, 0, image=self.photo, anchor=tk.NW)
        #self.canvas.pack(side=tk.LEFT)
        self.canvas.grid(row=0, column=1, rowspan=2, columnspan=4, sticky='nesw')

        # Add button: snapshot trigger
        self.snapshot_button = tk.Button(self.window,
                                         text='Query',
                                         command=self.snapshot)
        #self.snapshot_button.pack(expand=True)
        self.snapshot_button.grid(row=1, column=0)

        # Add button and label: threshold controller
        self.threshold = tk.Label(self.window,
                               text=self.crowd_factor,
                               font=('Courier New', 10),
                               justify=tk.LEFT)
        self.threshold.grid(row=1, column=1)

        self.snapshot_button = tk.Button(self.window,
                                         text='+',
                                         command=self.increase_threshold)
        self.snapshot_button.grid(row=1, column=2)

        self.snapshot_button = tk.Button(self.window,
                                         text='-',
                                         command=self.decrease_threshold)
        self.snapshot_button.grid(row=1, column=3)

        # Create data collector thread
        t = threading.Thread(name='Data Collector',
                             target=self.dc_service.run,
                             args=(self.dc_kwargs,))
        t.start()

        # Start the main UI program
        self.window.mainloop()

    def update(self, data, imgkey='bytes'):
        '''
        Args:
            data: Inference result loaded from JSON object
        '''
        # Retrieve result image
        jpg_bytes = payload.destringify_jpg(data[imgkey])
        img = payload.jpg2rgb(jpg_bytes)

        # Retrieve result text, and update text area
        data.pop(imgkey)
        result_text = self.process_output(data)
        if 'safely' in result_text:
            text_color = 'blue'
        else:
            text_color = 'red'
        self.result.config(text=result_text, fg=text_color)

        # update image area
        resized_img = Image.fromarray(img).resize((self.canvas_h, self.canvas_w))
        self.photo = ImageTk.PhotoImage(image=resized_img)
        win_w = self.photo.width() + self.result.winfo_width()
        win_h = self.photo.height() + self.snapshot_button.winfo_height()
        self.window.geometry('{}x{}'.format(win_w, win_h))
        self.canvas.itemconfig(self.image_id, image=self.photo)

    def snapshot(self):
        self.dc_service.send_snapshot_trigger()

    def increase_threshold(self):
        self.crowd_factor += 1
        self.threshold.config(text=self.crowd_factor)

    def decrease_threshold(self):
        self.crowd_factor -= 1
        self.threshold.config(text=self.crowd_factor)

    def process_output(self, output):
        '''
        Args:
            output: Inference result, JSON object

        Returns:
            Stringified JSON data.
        '''
        if 'annotations' in output.keys():
            count = 0
            for obj in output['annotations']:
                if obj['label'] == 'person':
                    count += 1
                #logger.info('label = {}'.format(k))
            msg = '{} persons at the corner\n\n'.format(count)
            if count > self.crowd_factor:
                msg += 'Too crowded,\nsuggest to go straight'
            else:
                msg += 'You can turn right safely'
            return msg
        else:
            return json.dumps(output, indent=4)

    def on_closing(self):
        self.dc_service.comm.disconnect()
        self.window.destroy()


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--data-dirpath',
        default=None,
        help='Dirpath where to store collected data.'
    )
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument(
        '--topic-config',
        default=None,
        help='Path of the MQTT topic subscription JSON.'
    )
    ap.add_argument(
        '--image-width',
        type=int,
        default=300,
        help='Image display width in pixel.'
    )
    ap.add_argument(
        '--image-height',
        type=int,
        default=300,
        help='Image display height in pixel.'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()

    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        }
    }
    dc_service = DataCollectorService(comm_config,
                                      args['data_dirpath'])
    UI(dc_service, args)


if __name__ == '__main__':
    main()



import argparse
import json
import logging

from os.path import join as pjoin

import cv2

from dyda.pipelines import pipeline as dydapl


logger = logging.getLogger('launcher')


class BerryNetPipelineLauncher(object):
    def __init__(self, config, dyda_config_path='',
                 output_dirpath='', verbosity=0, benchmark=False,
                 lab_flag=False):
        self.pipeline = dydapl.Pipeline(
            config,
            dyda_config_path=dyda_config_path,
            parent_result_folder=output_dirpath,
            verbosity=verbosity,
            lab_flag=lab_flag)
        self.benchmark = benchmark

    def run(self, bitmap, meta={}, base_name=None):
        """Run pipeline.

        Args:
            bitmap: Image data in BGR format (numpy array)

        Returns:
            Dictionary with contents or empty list.
        """
        self.pipeline.run(bitmap,
                          external_meta=meta,
                          benchmark=self.benchmark,
                          base_name=base_name)
        try:
            # You can get pipeline output(s) in two ways:
            #
            # 1. Simple (and single) output from pipeline.output.
            #    - 100% available
            #    - Data type is dynamic (it might be a JSON object, an image, etc.)
            #
            # 2. Multiple outputs from components defined in pipeline config.
            #    - Currently it is used in the scenario that
            #      you want to have JSON result and image simultaneously.
            #    - final_json_output and final_img_output are component name,
            #      not component types. We are considering to define them as standard.
            if all(key in self.pipeline.pipeline.keys()
                   for key in ['final_json_output', 'final_img_output']):
                logger.debug('Pipeline output type: multiple (launcher_berrynet)')
                output = {
                    'annotations': self.pipeline.pipeline['final_json_output']['output']['annotations'],
                    'bytes': self.pipeline.pipeline['final_img_output']['output']
                }
            else:
                logger.debug('Pipeline output type: simple (launcher_berrynet)')
                output = self.pipeline.output
        except Exception as e:
            logger.critical(e)
        return output


def get_args(argv=None):
    """ Prepare auguments for running the script. """

    parser = argparse.ArgumentParser(
        description='Pipeline.'
    )
    parser.add_argument(
        '-i', '--input',
        type=str,
        default=(
            '/home/shared/customer_data/acti/201711-ACTi-A/'
            '20171207_recording/acti_2017-12-07-1701/frame'),
        help='Input folder for ')
    parser.add_argument(
        '-o', '--output',
        type=str,
        default='/home/shared/DT42/test_data/'
        'test_auto_labeler_with_tracker/results/',
        help='Output folder for output_metadata')
    parser.add_argument(
        '--lab_flag',
        dest='lab_flag',
        action='store_true',
        default=False,
        help='True to enable related lab process.'
    )
    parser.add_argument(
        '-p', '--pipeline_config',
        type=str,
        default='/home/lab/dyda/pipeline.config',
        help='File contains the definition of pipeline application.'
    )
    parser.add_argument(
        '-t', '--dyda_config',
        type=str,
        default='',
        help='File contains the component definitions.'
    )
    parser.add_argument(
        "-v", "--verbosity",
        action="count",
        default=0,
        help="increase output verbosity"
    )
    return parser.parse_args(argv)


def main():
    """ Example for testing pipeline. """

    args = get_args()

    log_level = logging.WARNING
    if args.verbosity == 1:
        log_level = logging.INFO
    elif args.verbosity >= 2:
        log_level = logging.DEBUG
    formatter = logging.Formatter('[launcher] %(levelname)s %(message)s')
    console = logging.StreamHandler()
    console.setFormatter(formatter)
    logger.setLevel(log_level)
    logger.addHandler(console)

    logger.debug('lab_flag is %r' % args.lab_flag)

    pipeline = BerryNetPipelineLauncher(
        config=args.pipeline_config,
        dyda_config_path=args.dyda_config,
        output_dirpath=args.output,
        verbosity=args.verbosity,
        lab_flag=args.lab_flag
    )

    logger.debug('Running Reader and Selector for frames')
    source_dirpath = args.input
    input_number = 100
    for i in range(input_number):
        input_data = pjoin(source_dirpath, '00000{}.png'.format(570 + i))
        ext_data = cv2.imread(input_data)

        output_data = pipeline.run(ext_data)
        logger.debug('===== frame #{} ====='.format(i))
        logger.debug('input: {}'.format(input_data))
        if (len(output_data) > 0):
            with open('output_{}.json'.format(i), 'w') as f:
                json.dump(output_data, f, indent=4)


if __name__ == "__main__":
    main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Pipeline service with relay engine (default engine).
"""

import argparse
import json
import logging
import os
import sys
import time

from datetime import datetime

import paho.mqtt.publish as publish
import activation.verify
import numpy as np
import cv2

from berrynet import logger
from berrynet.comm import payload
from berrynet.engine import DLEngine
from berrynet.service import EngineService
from dt42lab.core import tools
from bntrainer.launcher_berrynet import BerryNetPipelineLauncher


class PipelineEngine(DLEngine):
    def __init__(self, config,
                 trainer_config_path='',
                 warmup_size=(480, 640, 3),
                 disable_warmup=False,
                 benchmark=False,
                 verbosity=0):
        self.launcher = BerryNetPipelineLauncher(
            config,
            trainer_config_path=trainer_config_path,
            verbosity=verbosity, benchmark=benchmark)
        self.pipeline_config = tools.parse_json(config)
        if not disable_warmup:
            self.warmup(shape=warmup_size)

    # def process_input(self, tensor):
    #     return tensor

    def inference(self, tensor, meta={}, base_name=None):
        """
        Args:
            tensor: Image data in BGR format (numpy array)

        Returns:
            Dictionary following generic inference spec and
            pipeline spec (by project).
        """
        return self.launcher.run(tensor, meta=meta, base_name=base_name)
        # return {
        #     'annotations': {
        #         'label': 'dt42',
        #         'confidence': 0.99
        #     }
        # }

    def process_output(self, output):
        return output

    def get_dl_component_config(self, pipeline_config):
        """Get pipeline def list containing only DL components

        Args:
            pipeline_config: pipeline config JSON object

        Returns:
            List of DL components definitions
        """
        dl_comp_config = []
        try:
            pipeline_def = pipeline_config['pipeline_def']
        except KeyError:
            logger.warning('Invalid pipeline config')
            pipeline_def = []
        for comp_config in pipeline_def:
            if ('classifier' in comp_config['name'] or
                    'detector' in comp_config['name']):
                dl_comp_config.append(comp_config)
        return dl_comp_config

    def warmup(self, shape=(480, 640, 3), iteration=5):
        """Warmup pipeline engine

        Use all-zero numpy array as input to warmup pipeline engine.

        Args:
            meta: Metadata of image data
            shape: Warmup image shape in (w, h, c) format
            iteration: How many times to feed in warmup image

        Returns:
            N/A
        """
        logger.debug('Warmup shape: {}'.format(shape))
        input_data = [np.zeros(shape=shape, dtype=np.uint8)] * iteration

        # FIXME: get engines programatically
        dl_comp_config = self.get_dl_component_config(self.pipeline_config)
        for comp_config in dl_comp_config:
            t_start = time.time()

            comp_name = comp_config['name']
            inst = self.launcher.pipeline.pipeline[comp_name]['instance']
            inst.input_data = input_data
            inst.main_process()

            t_duration = time.time() - t_start
            logger.debug('Warmup {0} costs {1} sec'.format(comp_name,
                                                           t_duration))


def duration(t):
    return (datetime.now() - t).microseconds / 1000


class PipelineDummyEngine(DLEngine):
    def inference(self, tensor, meta={}):
        output = None
        return output


class PipelineService(EngineService):
    def __init__(self, service_name, engine, comm_config,
                 pipeline_config_path=None,
                 disable_engine=False,
                 disable_warmup=False,
                 warmup_size=(480, 640, 3),
                 output_broker_ip="localhost",
                 output_mqtt_topic='berrynet/engine/pipeline/result'):
        super().__init__(service_name,
                         engine,
                         comm_config)

        self.pipeline_config_path = pipeline_config_path
        self.trainer_config_path = ''
        self.warmup_size = warmup_size

        self.disable_engine = disable_engine
        self.output_broker_ip = output_broker_ip
        self.output_mqtt_topic = output_mqtt_topic

        if not os.path.exists('/tmp/dlbox-pipeline'):
            os.mkdir('/tmp/dlbox-pipeline')
        self.counter = 0

        logger.debug('Pipeline result topic: {}'.format(
            self.output_mqtt_topic))

    def inference(self, pl):
        if self.disable_engine:
            self.dummy_inference(pl)
        else:
            self.dl_inference(pl)

    def dl_inference(self, pl):
        def empty_inference_result(count):
            return [
                {
                    'channel': i,
                    'annotations': []
                }
                for i in range(count)]

        t = datetime.now()
        base_name = None
        logger.debug('counter #{}'.format(self.counter))
        logger.debug('payload size: {}'.format(len(pl)))
        logger.debug('payload type: {}'.format(type(pl)))
        jpg_json = payload.deserialize_payload(pl.decode('utf-8'))
        jpg_bytes_list = [
            payload.destringify_jpg(img['bytes']) for img in jpg_json]
        metas = [img.get('meta', {}) for img in jpg_json]
        logger.debug('destringify_jpg: {} ms'.format(duration(t)))

        t = datetime.now()
        bgr_arrays = [
            payload.jpg2bgr(jpg_bytes) for jpg_bytes in jpg_bytes_list]
        logger.debug('jpg2bgr: {} ms'.format(duration(t)))

        t = datetime.now()
        image_data = self.engine.process_input(bgr_arrays)
        # FIXME: Galaxy pipeline doesn't support multiple metadata for multiple
        # images at the moment (which will be needed), so we provide the first
        # metadata here. This commit should be revert when Galaxy pipeline
        # support it: https://gitlab.com/DT42/galaxy42/dt42-trainer/issues/120
        meta_data = metas[0]
        try:
            logger.debug(meta_data)
            output = self.engine.inference(image_data,
                                           meta=meta_data,
                                           base_name=base_name)
            model_outputs = self.engine.process_output(output)
        except IndexError as e:
            # FIXME: workaround for pipeline
            # Pipeline throw IndexError when there's no results, see:
            # https://gitlab.com/DT42/galaxy42/dt42-trainer/issues/86
            # So we catch the exeception, and produce a dummy result
            # to hook. This workaround should be removed after the issue
            # has been fixed.
            model_outputs = empty_inference_result(len(jpg_json))
            logger.warning(('inference results are empty because '
                            'pipeline raised IndexError'))

        if model_outputs is None:
            model_outputs = empty_inference_result(len(jpg_json))
            logger.warning(('inference results are empty because '
                            'severe error happened in pipeline'))

        if isinstance(model_outputs, dict):
            model_outputs = [model_outputs]
        logger.debug('Result: {}'.format(model_outputs))
        logger.debug('Classification takes {} ms'.format(duration(t)))

        # self.engine.cache_data('model_output', model_outputs)
        # self.engine.cache_data('model_output_filepath', output_name)
        # self.engine.save_cache()

        self.send_result(self.generalize_result(jpg_json, model_outputs),
                         self.output_mqtt_topic)

        self.counter += 1

    def dummy_inference(self, pl):
        logger.debug('dummy_inference is called')

    def switch_mode(self, pl):
        """Switch pipeline service between inference and non-inference modes

        If Pipeline service receives berrynet/data/mode topic with
        "inference" in payload, service will switch to inference mode;

        If "idle" or "learning" in payload, service will switch to
        non-inference mode.

        Pipeline service will create pipeline engine and
        listen to specified topics only in inference mode.

        Args:
            pl: MQTT message payload
                valid value: {'inference', 'idle', 'learning'}

        Returns:
            N/A
        """
        mode = pl.decode('utf-8')
        if mode == 'inference':
            self.disable_engine = False
        else:
            self.disable_engine = True

    def deploy(self, pl):
        """Deploy newly retrained model for pipeline engine

        New trainer config filepath is in the payload.

        Args:
            pl: MQTT message payload w/ new trainer config filepath.

        Returns:
            N/A
        """
        trainer_config_path = pl.decode('utf-8')
        self.trainer_config_path = trainer_config_path
        self.comm.send('berrynet/data/deployed', '')
        logger.info(('New model has been deployed, '
                     'trainer config: {}'.format(self.trainer_config_path)))

    def generalize_result(self, eng_inputs, eng_outputs):
        # Pipeline returns None if any error happened
        if eng_outputs is None:
            eng_outputs = [{}]

        if len(eng_inputs) != len(eng_outputs):
            logger.warning('Input length != output length: {} != {}'.format(
                len(eng_inputs), len(eng_outputs)))
            # We guarantee len of inputs will always be 1 (at least now), so
            # it's safer to access eng_inputs by index than to eng_outputs
            c_id = int(eng_inputs[0]['meta']['channel_id'])
            eng_outputs = [eng_outputs[c_id]]

        # FIXME: Workaround for spec incompatibility
        # DLBox spec use 'image_blob', but BerryNet use 'bytes', so we have to
        # do a convert here
        for eng_in, eng_out in list(zip(eng_inputs, eng_outputs)):
            if isinstance(eng_out, np.ndarray):
                r, result_img = cv2.imencode('.jpg', eng_out)
                eng_in['bytes'] = payload.stringify_jpg(result_img)
            else:
                try:
                    eng_in.update(eng_out)
                except KeyError as e:
                    logger.exception(
                        '{} ({}): {}'.format(e.__class__, e.__doc__, e))
            eng_in['image_blob'] = eng_in.pop('bytes')
        return eng_inputs

    def send_result(
        self, generalized_result, topic='berrynet/engine/pipeline/result'):
        # NOTE: There are numpy float in pipeline output, so we use
        # tools.dump_json instead of payload.serialize_payload
        for r in generalized_result:
            try:
                c_id = r.get('channel', None) or r['meta']['channel_id']
                topic += '/{}'.format(c_id)
            except KeyError:
                logger.warn(
                    'No channel id found, set topic to {}'.format(topic))

            logger.debug('output topic: {}'.format(topic))
            publish.single(topic,
                           payload=tools.dump_json(generalized_result),
                           hostname=self.output_broker_ip)


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--pipeline-config',
                    help='File contains the definition '
                         'of pipeline application')
    ap.add_argument('--debug',
                    action='store_true',
                    help='Debug mode toggle')
    ap.add_argument('--benchmark',
                    action='store_true',
                    help='Benchmark mode toggle')
    ap.add_argument('--broker-ip', '--input-broker-ip',
                    default='localhost',
                    help='MQTT broker IP')
    ap.add_argument('--output-broker-ip',
                    default='localhost',
                    help='Result output MQTT broker IP')
    ap.add_argument('--output-mqtt-topic',
                    default='berrynet/engine/pipeline/result',
                    help='Result output MQTT topic')
    ap.add_argument('--topic-config',
                    default=None,
                    help='Path of the MQTT topic subscription JSON.')
    ap.add_argument('--topic',
                    nargs=2,
                    action='append',
                    default=None,
                    help=('Two params in "<topic> <handler>" format. '
                          'It can be declared multiple times.'))
    ap.add_argument('--disable-engine',
                    action='store_true',
                    help='Service disable engine initially')
    ap.add_argument('--disable-warmup',
                    action='store_true',
                    help='Skip warming up pipeline by black image')
    ap.add_argument('-v', '--verbosity',
                    action='count', default=0,
                    help='Output verbosity')
    ap.add_argument('-w', '--warmup-size',
                    nargs=2,
                    type=int,
                    default=(640, 480),
                    help='Warmup image\'s size, in format "w h", '
                         'e.g., "640 480"')
    return vars(ap.parse_args())


def main():
    result = activation.verify.check_license()
    if not result:
        sys.exit(1)

    # Process CLI arguments
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    topic_config['berrynet/data/mode'] = 'self.switch_mode'
    topic_config['berrynet/data/deploy'] = 'self.deploy'

    if args['topic'] is not None:
        for t, h in args['topic']:
            topic_config[t] = h

    w, h = args['warmup_size']
    # Setup pipeline service
    if args['disable_engine']:
        eng = PipelineDummyEngine()
    else:
        eng = PipelineEngine(args['pipeline_config'],
                             disable_warmup=args['disable_warmup'],
                             verbosity=args['verbosity'],
                             benchmark=args['benchmark'],
                             warmup_size=(h, w, 3))
    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': 1883
        }
    }
    engine_service = PipelineService(
        'pipeline service',
        eng,
        comm_config,
        pipeline_config_path=args['pipeline_config'],
        disable_engine=args['disable_engine'],
        disable_warmup=args['disable_warmup'],
        warmup_size=(h, w, 3),
        output_broker_ip=args['output_broker_ip'],
        output_mqtt_topic=args['output_mqtt_topic'])

    engine_service.run(args)


if __name__ == '__main__':
    main()



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Engine service is a bridge between incoming data and inference engine.
"""

import os
import time

from datetime import datetime

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


class EngineService(object):
    def __init__(self, service_name, engine, comm_config):
        self.service_name = service_name
        self.engine = engine
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = eval(functor)
        self.comm_config['subscribe']['berrynet/data/rgbimage'] = self.inference
        self.comm = Communicator(self.comm_config, debug=True)

    def inference(self, pl):
        duration = lambda t: (datetime.now() - t).microseconds / 1000

        t = datetime.now()
        logger.debug('payload size: {}'.format(len(pl)))
        logger.debug('payload type: {}'.format(type(pl)))
        jpg_json = payload.deserialize_payload(pl.decode('utf-8'))
        jpg_bytes = payload.destringify_jpg(jpg_json['bytes'])
        logger.debug('destringify_jpg: {} ms'.format(duration(t)))

        t = datetime.now()
        rgb_array = payload.jpg2rgb(jpg_bytes)
        logger.debug('jpg2rgb: {} ms'.format(duration(t)))

        t = datetime.now()
        image_data = self.engine.process_input(rgb_array)
        output = self.engine.inference(image_data)
        model_outputs = self.engine.process_output(output)
        logger.debug('Result: {}'.format(model_outputs))
        logger.debug('Classification takes {} ms'.format(duration(t)))

        #self.engine.cache_data('model_output', model_outputs)
        #self.engine.cache_data('model_output_filepath', output_name)
        #self.engine.save_cache()

        self.result_hook(self.generalize_result(jpg_json, model_outputs))

    def generalize_result(self, eng_input, eng_output):
        eng_input.update(eng_output)
        return eng_input

    def result_hook(self, generalized_result):
        logger.debug('base result_hook')

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.engine.create()
        self.comm.run()



#!/usr/bin/env python3
#
# Copyright 2019 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import io
import logging
import os
import tempfile
import tarfile
import time

import telegram.ext
from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


class TelegramBotService(object):
    def __init__(self, comm_config, token, target_label='', debug=False):
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = eval(functor)
        # NOTE: Maybe change the hard-coding topic to parameter in the future.
        self.comm_config['subscribe']['berrynet/data/rgbimage'] = self.single_shot
        self.comm = Communicator(self.comm_config, debug=True)
        if os.path.isfile(token):
            self.token = self.get_token_from_config(token)
        else:
            self.token = token
        self.target_label = target_label
        self.debug = debug

        # Telegram Updater employs Telegram Dispatcher which dispatches
        # updates to its registered handlers.
        self.updater = telegram.ext.Updater(self.token,
                                            use_context=True)
        self.cameraHandlers = []

        self.shot = False
        self.single_shot_chat_id = None

    def get_token_from_config(self, config):
        with open(config) as f:
            cfg = json.load(f)
        return cfg['token']

    def match_target_label(self, target_label, bn_result):
        labels = [r['label'] for r in bn_result['annotations']]
        if target_label in labels:
            logger.debug('Find {0} in inference result {1}'.format(target_label, labels))
            return True
        else:
            logger.debug('Not find {0} in inference result {1}'.format(target_label, labels))
            return False

    def update(self, pl):
        try:
            payload_json = payload.deserialize_payload(pl.decode('utf-8'))
            jpg_bytes = payload.destringify_jpg(payload_json["bytes"])
            jpg_file_descriptor = io.BytesIO(jpg_bytes)

            for u in self.cameraHandlers:
                if self.updater is None:
                    continue

                if self.target_label == '':
                    if len(payload_json['annotations']) > 0:
                        logger.debug("Send photo to %s" % u)
                        self.updater.bot.send_photo(chat_id = u, photo=jpg_file_descriptor)
                    else:
                        logger.debug("Does not detect any object, no action")
                elif self.match_target_label(self.target_label, payload_json):
                    logger.info("Send notification photo with result to %s" % u)
                    self.updater.bot.send_photo(chat_id = u, photo=jpg_file_descriptor)
                else:
                    pass
        except Exception as e:
            logger.info(e)

    def single_shot(self, pl):
        """Capture an image from camera client and send to the client.
        """
        if self.shot is True:
            try:
                payload_json = payload.deserialize_payload(pl.decode('utf-8'))
                # WORKAROUND: Support customized camera client.
                #
                # Original camera client sends an `obj` in payload,
                # Customized camera client sends an `[obj]` in payload.
                #
                # We are unifying the rules. Before that, checking the type
                # as workaround.
                if type(payload_json) is list:
                    logger.debug('WORDAROUND: receive and unpack [obj]')
                    payload_json = payload_json[0]
                jpg_bytes = payload.destringify_jpg(payload_json["bytes"])
                jpg_file_descriptor = io.BytesIO(jpg_bytes)

                logger.info('Send single shot')
                self.updater.bot.send_photo(chat_id=self.single_shot_chat_id,
                                            photo=jpg_file_descriptor)
            except Exception as e:
                logger.info(e)

            self.shot = False
        else:
            logger.debug('Single shot is disabled, do nothing.')

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.start_nb()
        self.connect_telegram(args)

    def connect_telegram(self, args):
        try:
            self.updater.dispatcher.add_handler(
                telegram.ext.CommandHandler('help', self.handler_help))
            self.updater.dispatcher.add_handler(
                telegram.ext.CommandHandler('hi', self.handler_hi))
            self.updater.dispatcher.add_handler(
                telegram.ext.CommandHandler('camera', self.handler_camera))
            self.updater.dispatcher.add_handler(
                telegram.ext.CommandHandler('stop', self.handler_stop))
            self.updater.dispatcher.add_handler(
                telegram.ext.CommandHandler('shot', self.handler_shot))
            if (args["has_getlog"]):
                self.updater.dispatcher.add_handler(
                    telegram.ext.CommandHandler('getlog', self.handler_getlog))
            self.updater.start_polling()
        except Exception as e:
            logger.critical(e)

    def handler_help(self, update, context):
        logger.info("Received command `help`")
        update.message.reply_text((
            'I support these commands:\n\n'
            'help - Display help message.\n'
            'hi - Test Telegram client.\n'
            'camera - Start camera.\n'
            'stop - Stop camera.\n'
            'shot - Take a shot from camera.'))

    def handler_hi(self, update, context):
        logger.info("Received command `hi`")
        update.message.reply_text(
            'Hi, {}'.format(update.message.from_user.first_name))

    def handler_camera(self, update, context):
        logger.info("Received command `camera`, chat id: %s" % update.message.chat_id)
        # Register the chat-id for receiving images
        if (update.message.chat_id not in self.cameraHandlers):
            self.cameraHandlers.append (update.message.chat_id)
        update.message.reply_text('Dear, I am ready to help send notification')

    def handler_stop(self, update, context):
        logger.info("Received command `stop`, chat id: %s" % update.message.chat_id)
        # Register the chat-id for receiving images
        while (update.message.chat_id in self.cameraHandlers):
            self.cameraHandlers.remove (update.message.chat_id)
        update.message.reply_text('Bye')

    def handler_shot(self, update, context):
        logger.info("Received command `shot`, chat id: %s" % update.message.chat_id)
        # Register the chat-id for receiving images
        self.shot = True
        self.single_shot_chat_id = update.message.chat_id
        logger.debug('Enable single shot.')

    def handler_getlog(self, update, context):
        logger.info("Received command `getlog`, chat id: %s" % update.message.chat_id)
        # Create temporary tar.xz file
        tmpTGZ1 = tempfile.NamedTemporaryFile(suffix=".tar.xz")
        tmpTGZ = tarfile.open(fileobj=tmpTGZ1, mode="w:xz")
        tmpTGZPath = tmpTGZ1.name

        # Traverse /var/log
        varlogDir = os.path.abspath(os.path.join(os.sep, "var", "log"))
        for root, dirs, files in os.walk(varlogDir):
            for file in files:
                fullPath = os.path.join(root, file)
                # Check if the file is a regular file
                if not os.path.isfile(fullPath):
                    continue
                # Check if the file is accessable
                if not os.access(fullPath, os.R_OK):
                    continue
                # Pack the file
                tmpTGZ.add(name = fullPath, recursive=False)
        tmpTGZ.close()
        self.updater.bot.send_document(chat_id = update.message.chat_id, document = open(tmpTGZPath, 'rb'), filename=time.strftime('berrynet-varlog_%Y%m%d_%H%M%S.tar.xz'))

def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--token',
        help=('Telegram token got from BotFather, '
              'or filepath of a JSON config file with token.')
    )
    ap.add_argument(
        '--target-label',
        default='',
        help='Send a notification if the target label is in the result.'
    )
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument(
        '--topic',
        nargs='*',
        default=['berrynet/engine/tflitedetector/result'],
        help='The topic to listen, and can be indicated multiple times.'
    )
    ap.add_argument(
        '--topic-action',
        default='self.update',
        help='The action for the indicated topics.'
    )
    ap.add_argument(
        '--topic-config',
        default=None,
        help='Path of the MQTT topic subscription JSON.'
    )
    ap.add_argument('--debug',
        action='store_true',
        help='Debug mode toggle'
    )
    ap.add_argument('--has-getlog',
        action='store_true',
        help='Enable getlog command'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()
    if args['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # Topics and actions can come from two sources: CLI and config file.
    # Setup topic_config by parsing values from the two sources.
    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    topic_config.update({t:args['topic_action'] for t in args['topic']})

    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        }
    }
    telbot_service = TelegramBotService(comm_config,
                                        args['token'],
                                        args['target_label'],
                                        args['debug'])
    telbot_service.run(args)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

import os

import cv2
import numpy as np

from mvnc import mvncapi as mvnc


class MovidiusNeuralGraph(object):
    def __init__(self, graph_filepath, label_filepath):
        # mvnc.SetGlobalOption(mvnc.GlobalOption.LOGLEVEL, 2)
        devices = mvnc.EnumerateDevices()
        if len(devices) == 0:
            raise Exception('No devices found')
        self.device = mvnc.Device(devices[0])
        self.device.OpenDevice()

        # Load graph
        with open(graph_filepath, mode='rb') as f:
            graphfile = f.read()
        self.graph = self.device.AllocateGraph(graphfile)

        # Load labels
        self.labels = []
        with open(label_filepath, 'r') as f:
            for line in f:
                label = line.split('\n')[0]
                if label != 'classes':
                    self.labels.append(label)
            f.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.graph.DeallocateGraph()
        self.device.CloseDevice()

    def inference(self, data):
        self.graph.LoadTensor(data.astype(np.float16), 'user object')
        output, userobj = self.graph.GetResult()
        return output

    def get_graph(self):
        return self.graph

    def get_labels(self):
        return self.labels


def process_inceptionv3_input(img):
    image_size = 299
    mean = 128
    std = 1.0/128

    dx, dy, dz = img.shape
    delta = float(abs(dy - dx))
    if dx > dy:  # crop the x dimension
        img = img[int(0.5*delta):dx-int(0.5*delta), 0:dy]
    else:
        img = img[0:dx, int(0.5*delta):dy-int(0.5*delta)]
    img = cv2.resize(img, (image_size, image_size))
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    for i in range(3):
        img[:, :, i] = (img[:, :, i] - mean) * std
    return img


def process_inceptionv3_output(output, labels):
    processed_output = {'annotations': []}
    decimal_digits = 2
    top_k = 5
    top_inds = output.argsort()[::-1][:top_k]
    for i in range(top_k):
        human_string = labels[top_inds[i]]
        score = round(float(output[top_inds[i]]), decimal_digits)
        anno = {
            'type': 'classification',
            'label': human_string,
            'confidence': score
        }
        processed_output['annotations'].append(anno)
    return processed_output
    #return [(labels[top_inds[i]], output[top_inds[i]]) for i in range(5)]


def print_inceptionv3_output(output, labels):
    top_inds = output.argsort()[::-1][:5]

    for i in range(5):
        print(top_inds[i], labels[top_inds[i]], output[top_inds[i]])


def process_mobilenetssd_input(bgr_nparray):
    """Normalize MobileNet SSD input image.

    Args:
        img: Image nparray in BGR color model.

    Returns:
        Pre-processed image nparray in BGR color model.
    """
    img = cv2.resize(bgr_nparray, (300, 300))
    img = img - 127.5
    img = img * 0.007843
    return img


def process_mobilenetssd_output(output, img_w, img_h, labels, threshold=0.1):
    """
    More details about inference result format:
    https://github.com/movidius/ncappzoo/blob/master/caffe/SSD_MobileNet/run.py

    Args:
        output: Inference result returned by Graph.GetResult().
        img_w: Width of input image.
        img_h: Height of input image.
        labels:
        threshold:

    Returns:
        Annotations as dictionary, key is "annotations" and
        value a list of parsed results.

        Example:

            'annotations': [
                {
                    "label": "car",
                    "confidence": 0.93,
                    "left": 100,
                    "top": 100,
                    "right": 200,
                    "bottom": 200
                },
                ...
            ]
    """
    boxnum_index = 0
    result_index = 7
    result_size = 7
    num_valid_boxes = int(output[boxnum_index])
    annotations = []

    for i in range(num_valid_boxes):
        base_index = result_index + result_size * i
        result_objinfo = output[base_index:(base_index + result_size)]

        anno = {}
        anno['label'] = labels[int(result_objinfo[1])]
        anno['confidence'] = float(result_objinfo[2])
        anno['left'] = int(result_objinfo[3] * img_w)
        anno['top'] = int(result_objinfo[4] * img_h)
        anno['right'] = int(result_objinfo[5] * img_w)
        anno['bottom'] = int(result_objinfo[6] * img_h)

        if anno['confidence'] >= threshold:
            annotations.append(anno)

    return {'annotations': annotations}


if __name__ == '__main__':
    graph_filepath = ''  # model filepath
    label_filepath = ''  # label filepath
    path_to_images = ''  # image dirpath
    image_filenames = [os.path.join(path_to_images, image_name)
                       for image_name in []]  # image filename list

    movidius = MovidiusNeuralGraph(graph_filepath, label_filepath)
    labels = movidius.get_labels()

    print(''.join(['*' for i in range(79)]))
    print('inception-v3 on NCS')
    for image_filename in image_filenames:
        img = cv2.imread(image_filename).astype(np.float32)
        img = process_inceptionv3_input(img)
        print(''.join(['*' for i in range(79)]))
        print('Start download to NCS...')
        output = movidius.inference(img)
        print_inceptionv3_output(output, labels)

    print(''.join(['*' for i in range(79)]))
    print('Finished')



# Copyright 2018 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Data collector service.
"""

import argparse
import json
import os

from datetime import datetime
from os.path import join as pjoin

from berrynet import logger
from berrynet.comm import Communicator
from berrynet.comm import payload


class DataCollectorService(object):
    def __init__(self, comm_config, data_dirpath):
        self.comm_config = comm_config
        for topic, functor in self.comm_config['subscribe'].items():
            self.comm_config['subscribe'][topic] = eval(functor)
        self.comm_config['subscribe']['berrynet/engine/tensorflow/result'] = self.update
        self.comm_config['subscribe']['berrynet/engine/mvclassification/result'] = self.update
        self.comm = Communicator(self.comm_config, debug=True)
        self.data_dirpath = data_dirpath

    def update(self, pl):
        if not os.path.exists(self.data_dirpath):
            try:
                os.mkdir(self.data_dirpath)
            except Exception as e:
                logger.warn('Failed to create {}'.format(self.data_dirpath))
                raise(e)

        payload_json = payload.deserialize_payload(pl.decode('utf-8'))
        jpg_bytes = payload.destringify_jpg(payload_json['bytes'])
        payload_json.pop('bytes')
        logger.debug('inference text result: {}'.format(payload_json))

        timestamp = datetime.now().isoformat()
        with open(pjoin(self.data_dirpath, timestamp + '.jpg'), 'wb') as f:
            f.write(jpg_bytes)
        with open(pjoin(self.data_dirpath, timestamp + '.json'), 'w') as f:
            f.write(json.dumps(payload_json, indent=4))

    def save_pipeline_result(self, pl):
        if not os.path.exists(self.data_dirpath):
            try:
                os.mkdir(self.data_dirpath)
            except Exception as e:
                logger.warn('Failed to create {}'.format(self.data_dirpath))
                raise(e)

        payload_json = payload.deserialize_payload(pl.decode('utf-8'))
        jpg_bytes = payload.destringify_jpg(payload_json['image_blob'])
        payload_json.pop('image_blob')
        logger.debug('inference text result: {}'.format(payload_json))

        timestamp = datetime.now().isoformat()
        with open(pjoin(self.data_dirpath, timestamp + '.jpg'), 'wb') as f:
            f.write(jpg_bytes)
        with open(pjoin(self.data_dirpath, timestamp + '.json'), 'w') as f:
            f.write(json.dumps(payload_json, indent=4))

    def run(self, args):
        """Infinite loop serving inference requests"""
        self.comm.run()


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--data-dirpath',
        default='/tmp/berrynet-data',
        help='Dirpath where to store collected data.'
    )
    ap.add_argument(
        '--broker-ip',
        default='localhost',
        help='MQTT broker IP.'
    )
    ap.add_argument(
        '--broker-port',
        default=1883,
        type=int,
        help='MQTT broker port.'
    )
    ap.add_argument(
        '--topic-config',
        default=None,
        help='Path of the MQTT topic subscription JSON.'
    )
    return vars(ap.parse_args())


def main():
    args = parse_args()

    if args['topic_config']:
        with open(args['topic_config']) as f:
            topic_config = json.load(f)
    else:
        topic_config = {}
    comm_config = {
        'subscribe': topic_config,
        'broker': {
            'address': args['broker_ip'],
            'port': args['broker_port']
        }
    }
    dc_service = DataCollectorService(comm_config,
                                      args['data_dirpath'])
    dc_service.run(args)


if __name__ == '__main__':
    main()



# Copyright 2017 DT42
#
# This file is part of BerryNet.
#
# BerryNet is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BerryNet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BerryNet.  If not, see <http://www.gnu.org/licenses/>.

"""Darkflow inference engine.
"""

from __future__ import print_function

import argparse
import logging

import cv2

from darkflow.net.build import TFNet
from engineservice import EngineService
from engineservice import DLEngine
from dlmodelmgr import DLModelManager


# FIXME: Make these variables configurable
SystemSnapshot = '/usr/local/berrynet/dashboard/www/freeboard/snapshot.jpg'


class DarkflowEngine(DLEngine):
    def __init__(self, model, label, config):
        super(DarkflowEngine, self).__init__()
        self.engine_options = {
            'model': config,
            'load': model,
            #'model': "cfg/tiny-yolo.cfg",
            #'load': "bin/tiny-yolo.weights",
            'verbalise': True,
            "threshold": 0.1
        }

    def create(self):
        self.tfnet = TFNet(self.engine_options)

    def inference(self, tensor):
        return self.tfnet.return_predict(tensor)

    def save_cache(self):
        #with open(self.cache['model_output_filepath'], 'w') as f:
        #    f.write(str(self.cache['model_output']))
        drawBoundingBoxes(self.cache['model_input'],
                          #self.cache['model_output_filepath'] + '.jpg',
                          SystemSnapshot,
                          self.cache['model_output'],
                          self.tfnet.meta['colors'])


def drawBoundingBoxes(imageData, imageOutputPath, inferenceResults, colorMap):
    """Draw bounding boxes on an image.

    imageData: image data in numpy array format
    imageOutputPath: output image file path
    inferenceResults: Darkflow inference results
    colorMap: Bounding box color candidates, list of RGB tuples.
    """
    # TODO: return raw data instead of save image
    for res in inferenceResults:
        left = res['topleft']['x']
        top = res['topleft']['y']
        right = res['bottomright']['x']
        bottom = res['bottomright']['y']
        colorIndex = res['coloridx']
        color = colorMap[colorIndex]
        label = res['label']
        confidence = res['confidence']
        imgHeight, imgWidth, _ = imageData.shape
        thick = int((imgHeight + imgWidth) // 300)

        cv2.rectangle(imageData,(left, top), (right, bottom), color, thick)
        cv2.putText(imageData, label, (left, top - 12), 0, 1e-3 * imgHeight,
            color, thick//3)
    cv2.imwrite(imageOutputPath, imageData)
    logging.debug('Save bounding box result image to {}'.format(imageOutputPath))


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('--model',
                    help='Model file path')
    ap.add_argument('--label',
                    help='Label file path')
    ap.add_argument('--model_package',
                    default='',
                    help='Model package "name-version" naming')
    ap.add_argument('--image_dir', required=True,
                    help='Path to image file')
    ap.add_argument('--service_name', required=True,
                    help='Engine service name used as PID filename')
    ap.add_argument('--num_top_predictions', default=5,
                    help='Display this many predictions')
    return vars(ap.parse_args())


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    args = parse_args()
    if args['model_package'] != '':
        dlmm = DLModelManager()
        meta = dlmm.get_model_meta(args['model_package'])
        args['model'] = meta['model']
        args['label'] = meta['label']
        args['config'] =  meta['config']['graph']
    logging.debug('model filepath: ' + args['model'])
    logging.debug('label filepath: ' + args['label'])
    logging.debug('image_dir: ' + args['image_dir'])

    darkflow_engine = DarkflowEngine(args['model'], args['label'], args['config'])
    engine_service = EngineService(args['service_name'], darkflow_engine)
    engine_service.run(args)

    # this code block works
    #import cv2
    #input_tensor = cv2.imread('/tmp/berrynet/dog.jpg')
    #tensor = darkflow_engine.process_input(input_tensor)
    #output = darkflow_engine.inference(tensor)
    #output = darkflow_engine.process_output(output)



"""

$python xmlTotxt.py -n $FOLDER -u $LABELME_USER
 --classes $CLASS_1 $CLASS_2 $CLASS_3...

FOLDER: LabelMe project folder name
LABELME_USER: LabelMe user name
CLASS_i: Class labels defined on LabelMe

"""
import argparse
import logging
import os
import xml.etree.ElementTree as ET

from os.path import join
from shutil import copyfile


def convert(size, in_x, in_y):
    dw = 1./size[0]
    dh = 1./size[1]
    x = (in_x[0] + in_x[1])/2.0
    y = (in_y[0] + in_y[1])/2.0
    w = in_x[1] - in_x[0]
    h = in_y[1] - in_y[0]
    x = x*dw
    w = w*dw
    y = y*dh
    h = h*dh
    return (x, y, w, h)


def convert_annotation(in_dir, out_dir, image_id, out_id, classes):
    in_file = open("%s/%s.xml" % (in_dir, image_id))
    o_file = open(out_dir + "/%s.txt" % out_id, "w")
    tree = ET.parse(in_file)
    root = tree.getroot()
    size = root.find("imagesize")
    h = float(size.find("nrows").text)
    w = float(size.find("ncols").text)

    for obj in root.iter("object"):
        X = []
        Y = []
        cls = obj.find("name").text
        if cls not in classes:
            logging.debug("%s is not in the selected class" % cls)
            continue
        cls_id = classes.index(cls)
        for pt in obj.find("polygon").findall("pt"):
            X.append(float(pt.find("x").text))
            Y.append(float(pt.find("y").text))
        if (len(X) < 2 or len(Y) < 2):
            logging.warning("%s doesn't have sufficient info, ignore" % cls)
            continue
        X = list(set(X))
        X.sort()
        Y = list(set(Y))
        Y.sort()
        bb = convert((w, h), X, Y)
        o_file.write(str(cls_id) + " " + " ".join([str(a) for a in bb]) + "\n")


def find_output_id(out_img_dir, image_id, suffix):
    counter = 0
    out_img_path = join(out_img_dir, image_id + "." + suffix)
    out_id = image_id
    while os.path.exists(out_img_path):
        logging.info("%s exists" % out_img_path)
        counter += 1
        out_id = image_id + "_" + str(counter)
        out_img_path = join(out_img_dir, out_id + "." + suffix)
    return out_id, out_img_path


def main():

    parser = argparse.ArgumentParser(
        description="Simple tool to make the scene image better."
        )
    parser.add_argument(
        "-v", "--verbosity", action="count",
        help="increase output verbosity"
        )
    parser.add_argument(
        "-r", "--root", type=str, default=None,
        help="Specify the root directory (default: PWD)"
        )
    parser.add_argument(
        "-n", "--name", type=str, default="youtube_09",
        help="Specify the name of the original video (default: youtube_09)"
        )
    parser.add_argument(
        "-u", "--user", type=str, default="V",
        help="Specify the username (default: V)"
        )
    parser.add_argument(
        "-o", "--outdir", type=str, default="test2017",
        help="Output dir in root, must end with 2017 (default: test2017)"
        )
    parser.add_argument(
        "--classes", nargs="+", type=str, default=["fighting", "dog"],
        help="Classes to be trained. Default: [fighting, dog]"
        )
    parser.add_argument(
        "-d", "--delete", action="store_true",
        help="Use this option to clean up train.txt"
        )

    args = parser.parse_args()

    log_level = logging.WARNING
    if args.verbosity == 1:
        log_level = logging.INFO
    elif args.verbosity >= 2:
        log_level = logging.DEBUG
    logging.basicConfig(level=log_level,
                        format="[xmlTotxt: %(levelname)s] %(message)s")

    logging.info(args.classes)

    if args.root is None:
        args.root = os.getcwd()
    in_dir = join(args.root, "Annotations", "users",
                  args.user, args.name)
    in_img_dir = join(args.root, "Images", "users",
                      args.user, args.name)
    out_dir = join(args.root, args.outdir, "labels")
    out_img_dir = join(args.root, args.outdir, "JPEGImages")

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    if not os.path.exists(out_img_dir):
        os.makedirs(out_img_dir)

    foutput = join(args.root, "train.txt")
    if args.delete:
        output = open(foutput, "w")
    else:
        output = open(foutput, "a")

    for _img_path in os.listdir(in_img_dir):
        img_path = os.path.join(in_img_dir, _img_path)
        suffix = os.path.basename(_img_path).split(".")[1]
        logging.debug("Find %s" % img_path)
        image_id = os.path.basename(_img_path).split(".")[0]
        out_id, out_img_path = find_output_id(out_img_dir, image_id, suffix)
        convert_annotation(in_dir, out_dir, image_id, out_id, args.classes)
        logging.info("Copy %s to %s" % (img_path, out_img_path))
        copyfile(img_path, out_img_path)

        output.write(out_img_path + "\n")


if __name__ == '__main__':
    main()
